#
# Copyright 2014-2015 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

'''AIV Recipes for EMIR'''

from __future__ import division, print_function

import math

import numpy as np
from scipy.interpolate import splrep, splev, sproot
from photutils import CircularAperture
from photutils import aperture_photometry
from photutils.geometry.circular_overlap import circular_overlap_grid

from astropy.modeling import (fitting, models)

from numina.array.mode import mode_half_sample
from numina.array.recenter import centering_centroid
from numina.array.utils import wcs_to_pix_np
from numina.array.fwhm import compute_fwhm_1d_simple
from numina.array.utils import image_box2d
from numina.modeling import EnclosedGaussian
from numina.constants import FWHM_G


def encloses_annulus(x_min, x_max, y_min, y_max, nx, ny, r_in, r_out):
    '''Encloses function backported from old photutils'''

    gout = circular_overlap_grid(x_min, x_max, y_min, y_max, nx, ny, r_out, 1, 1)
    gin = circular_overlap_grid(x_min, x_max, y_min, y_max, nx, ny, r_in, 1, 1)
    return gout - gin


def comp_back_with_annulus(img, xc, yc, r_in, r_out, frac=0.1):
    '''
    center: [x,y], center of first pixel is [0,0]
    '''

    x_min = -0.5 - xc
    x_max = img.shape[1] - 0.5 - xc
    y_min = -0.5 - yc
    y_max = img.shape[1] - 0.5 - yc
    mm = encloses_annulus(x_min, x_max, y_min, y_max,
                          img.shape[1], img.shape[0],
                          r_in, r_out)

    valid = mm > frac
    rr = img[valid]

    if rr.size == 0:
        raise ValueError("Not enough points to compute background")

    # mode?
    bck = mode_half_sample(rr)

    return bck, mm


class ConstantBackgroundEstimator(object):
    def __init__(self, const):
        self.const = const

    def __call__(self, a, x, y):
        return self.const


class AnnulusBackgroundEstimator(object):
    def __init__(self, r1, r2):
        self.r1 = r1
        self.r2 = r2

    def __call__(self, a, x, y):
        bck, _ = comp_back_with_annulus(a, x, y, self.r1, self.r2)
        return bck


def compute_fwhm_enclosed(imgs, xc, yc, minrad=0.01, maxrad=15.0):

    peak_pix = wcs_to_pix_np((xc, yc))
    peak = imgs[tuple(peak_pix)]

    rad = np.logspace(np.log10(minrad), np.log10(maxrad), num=100)
    flux = np.zeros_like(rad)
    positions = [(xc, yc)]
    for idr, r in enumerate(rad):
        ca = CircularAperture(positions, r)
        m = aperture_photometry(imgs, ca)
        flux[idr] = m['aperture_sum'][0]

    idx = flux.argmax()

    rmodel = rad[:idx+1]
    fmodel = flux[:idx+1]
    fmax = fmodel[-1]

    res_d = fit_fwhm_enclosed_direct(peak, rad, flux)

    res_g = fit_fwhm_enclosed_grow(fmax, rmodel, fmodel)

    return res_d, res_g


def compute_fwhm_enclosed_direct(imgs, xc, yc, minrad=0.01, maxrad=15.0):

    peak_pix = wcs_to_pix_np((xc, yc))
    peak = imgs[tuple(peak_pix)]

    rad = np.logspace(np.log10(minrad), np.log10(maxrad), num=100)
    flux = np.zeros_like(rad)
    positions = [(xc, yc)]
    for idr, r in enumerate(rad):
        ca = CircularAperture(positions, r)
        m = aperture_photometry(imgs, ca)
        flux[idr] = m['aperture_sum'][0]

    return fit_fwhm_enclosed_direct(peak, rad, flux)


def fit_fwhm_enclosed_direct(peak, rad, flux):

    # We use splines to interpolate and derivate
    spl = splrep(rad, flux)
    # First derivative
    vald1 = splev(rad, spl, der=1)

    splinter = splrep(rad, vald1 - math.pi * peak * rad)

    roots = sproot(splinter)
    nroots = len(roots)
    if peak < 0:
        msg = "The method doesn't converge, peak is negative"
        fwhm = -99
    else:
        if nroots == 0:
            msg = "The method doesn't converge, no roots"
            fwhm = -99
        elif nroots == 1:
            r12 = roots[0]
            fwhm = 2 * r12
            msg = "The method converges, one root"
        else:
            msg = "The method doesn't converge, multiple roots"
            r12 = roots[0]
            fwhm = 2 * r12

    return peak, fwhm, msg


def compute_fwhm_enclosed_grow(imgs, xc, yc, minrad=0.01, maxrad=15.0):

    rad = np.logspace(np.log10(minrad), np.log10(maxrad), num=100)
    flux = np.zeros_like(rad)
    positions = [(xc, yc)]
    for idr, r in enumerate(rad):
        ca = CircularAperture(positions, r)
        m = aperture_photometry(imgs, ca)
        flux[idr] = m['aperture_sum'][0]
    idx = flux.argmax()
    rmodel = rad[:idx+1]
    fmodel = flux[:idx+1]
    fmax = fmodel[-1]

    return fit_fwhm_enclosed_grow(fmax, rmodel, fmodel)


def fit_fwhm_enclosed_grow(fmax, rad, flux):

    fitter = fitting.LevMarLSQFitter()
    model1 = EnclosedGaussian(amplitude=fmax, stddev=1.0)
    result = fitter(model1, rad, flux)

    amplitude = result.amplitude.value
    sigma = result.stddev.value
    fwhm = sigma * FWHM_G
    peak = amplitude / (2 * math.pi * sigma * sigma)

    return amplitude, fwhm, peak, "done"


def moments(data, x0, y0, half_box):

    sl1 = image_box2d(x0, y0, data.shape, half_box)

    part1 = data[sl1]
    norm = part1.sum()

    Y1, X1 = np.mgrid[sl1]

    xo = X1 - x0
    yo = Y1 - y0

    Mxx = (xo**2 * part1).sum() / norm
    Myy = (yo**2 * part1).sum() / norm
    Mxy = (xo * yo * part1).sum() / norm
    e = math.sqrt((Mxx - Myy)**2 + (2 * Mxy)**2) / (Mxx + Myy)
    pa = 0.5 * math.atan(2 * Mxy / (Mxx - Myy))

    return Mxx, Myy, Mxy, e, pa


def rim(data, xinit, yinit,
        recenter_half_box=5,
        recenter_nloop=10,
        recenter_maxdist=10.0, buff=3, width=5, niter=10, rplot=8):

    fine_recentering = False
    sigma0 = 1.0
    rad = 3 * sigma0 * FWHM_G
    box = recenter_half_box
    recenter_half_box = (box, box)
    plot_half_box = (3*box, 3*box)

    print('C initial', xinit, yinit)
    x0, y0, _1, _2, _3 = centering_centroid(data, xinit, yinit,
                                            box=recenter_half_box,
                                            maxdist=recenter_maxdist,
                                            nloop=recenter_nloop
                                            )
    print('C final', x0, y0)

    if fine_recentering:
        print('Fine recentering')
        print('C initial', x0, y0)
        x1, y1, _back, _status, _msg = centering_centroid(
            data,
            x0,
            y0,
            box=(1, 1),
            maxdist=2*math.sqrt(2),
            nloop=1
            )
        print('C final', x1, y1)

    sl = image_box2d(x0, y0, data.shape, plot_half_box)
    part = data[sl]

    xx0 = x0 - sl[1].start
    yy0 = y0 - sl[0].start

    Y, X = np.mgrid[sl]

    # Photometry
    D = np.sqrt((X-x0)**2 + (Y-y0)**2)
    m = D < rplot
    r1 = D[m]

    fitter = fitting.LevMarLSQFitter()
    model = models.Gaussian1D(amplitude=1.0, mean=0, stddev=1.0)
    model.mean.fixed = True  # Mean is always 0.0

    irad = rad
    for i in range(niter):
        rs1 = rad + buff
        rs2 = rs1 + width

        bckestim = AnnulusBackgroundEstimator(r1=rs1, r2=rs2)
        bck = bckestim(part, xx0, yy0)
        part_s = part - bck

        ca = CircularAperture([(xx0, yy0)], rad)
        m = aperture_photometry(part_s, ca)
        flux_aper = m['aperture_sum'][0]

        f1 = part_s[m]
        g1d_f = fitter(model, r1, f1, weights=(r1+1e-12)**-1)

        rpeak = g1d_f.amplitude.value
        # sometimes the fit is negative
        rsigma = abs(g1d_f.stddev.value)

        rfwhm = rsigma * FWHM_G

        dpeak, dfwhm, smsg = compute_fwhm_enclosed_direct(part_s, xx0, yy0)

        rad = 3 * dfwhm
        if abs(rad-irad) < 1e-3:
            # reached convergence
            print('convergence in iter %d' % (i+1))
            break
        else:
            irad = rad
    else:
        print('no convergence in photometric radius determination')

    print('P, aper rad', rad, 'flux_aper', flux_aper[0])
    print('P, annulus background:', bck, 'radii', rs1, rs2)
    eamp, efwhm, epeak, emsg = compute_fwhm_enclosed_grow(
        part_s, xx0, yy0, maxrad=rs1
        )
    print('Enclosed fit, peak:', epeak, 'fwhm', efwhm)
    print('Radial fit, peak:', rpeak, 'fwhm', rfwhm)
    print('Direct enclosed, peak:', dpeak, 'dfwhm', dfwhm)

    lpeak, fwhm_x, fwhm_y = compute_fwhm_1d_simple(part_s, xx0, yy0)
    print('Simple, peak:', lpeak, 'fwhm x', fwhm_x, 'fwhm y', fwhm_y)

    # Fit in a smaller box
    fit2d_rad = int(math.ceil(0.5 * rad))

    fit2d_half_box = (fit2d_rad, fit2d_rad)
    sl1 = image_box2d(x0, y0, data.shape, fit2d_half_box)
    part1 = data[sl1]
    Y1, X1 = np.mgrid[sl1]
    g2d = models.Gaussian2D(amplitude=rpeak, x_mean=x0, y_mean=y0,
                            x_stddev=1.0, y_stddev=1.0)
    g2d_f = fitter(g2d, X1, Y1, part1 - bck)
    print('Gauss2D fit')
    print(g2d_f)

    moments_half_box = fit2d_half_box
    Mxx, Myy, Mxy, e, pa = moments(data, x0, y0, moments_half_box)

    print(Mxx, Myy, Mxy, e, pa)




import pytest

from ..dtu_configuration import DtuConfiguration


@pytest.mark.xfail(reason="DtuConfiguration.__eq__ not complete")
def test_dtu_eq():

    dtu1 = DtuConfiguration()
    assert not (dtu1 == 100)


@pytest.mark.xfail(reason="DtuConfiguration.__ne__ not complete")
def test_dtu_eq():

    dtu1 = DtuConfiguration()
    assert dtu1 != 100


def compute_md5_checksum(conf):
    import hashlib
    mm = hashlib.md5()
    ndig = 3
    values = (conf.xdtu, conf.ydtu, conf.zdtu, conf.xdtu_0, conf.ydtu_0, conf.zdtu_0)
    m1 = [round(r, ndig) for r in values]
    m2 = ["{:+010.3f}".format(m) for m in m1]
    m3 = ':'.join(m2)
    mm.update(m3.encode('utf-8'))
    return mm.hexdigest()


def test_checksum():
    import itertools

    dtu1 = DtuConfiguration.define_from_values(
        -199.871994018555, -6.01818990707397, -535.546997070312,
        -200.02099609375, -5.93765020370483, -535.0
    )

    dtu2 = DtuConfiguration.define_from_values(
        -199.871994018555, -6.01818990707397, -535.546997070312,
        -200.02099609375, -5.93765020370483, -535.0
    )

    dtu3 = DtuConfiguration.define_from_values(
        -199.872, -6.01818990707397, -535.546997070312,
        -200.02099609375, -5.93765020370483, -535.0
    )

    dtu4 = DtuConfiguration.define_from_values(
        -100, -6, 535.546111,
        -1200.0, -5.937650, -535.0
    )

    for a, b in itertools.combinations([dtu1, dtu2, dtu3, dtu4], 2):
        ca = compute_md5_checksum(a)
        cb = compute_md5_checksum(b)
        if a == b:
            assert ca == cb
        else:
            assert ca != cb



#
# Copyright 2008-2019 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import print_function

from math import sin, cos, tan, pi
import pkgutil

from six import StringIO

from scipy import array
import scipy.interpolate as sil
from scipy import loadtxt


# Classes are new style
__metaclass__ = type


class Counts:
    INTEGRAL_COUNTS = 0,
    DIFFERENTIAL_COUNTS = 1


class PhotometricFilter:
    FILTER_NONE = 0
    FILTER_B = 1
    FILTER_V = 2
    FILTER_Z = 3
    FILTER_J = 4
    FILTER_H = 5
    FILTER_K = 6


class RBModel:
    """Star counts model from Ratnatunga & Bahcall.

    Number of stars per square arc minute
    """
    _counts = array(
        [
            [12, 14, 16, 18, 20, 22, 24, 26, 28],
            [0.0, 1.3e-2, 1.8e-2, 4.8e-2, 1.1e-1,
                9.0e-2, 2.7e-2, 3.4e-3, 1.6e-4],
            [0.0, 8.2e-3, 2.6e-2, 2.6e-2, 4.5e-2,
                1.2e-1, 1.1e-1, 3.9e-2, 5.6e-3],
            [0.0, 2.0e-3, 1.0e-2, 4.4e-2, 1.1e-1,
                2.2e-1, 4.6e-1, 8.1e-1, 1.1e-0]
        ]
    )
    # Order 3 spline
    _spl = [sil.splrep(_counts[0], _counts[i]) for i in (1, 2, 3)]
    _min = _counts[0, 0]
    del _counts

    _null = [0.0, 0.0, 0.0]
    _colorsVK = [4.7, 2.5, 1.4]

    name = 'Ratnatunga & Bahcall'
    filters = [PhotometricFilter.FILTER_K]

    def differential_counts(self, mag, filter=PhotometricFilter.FILTER_K):
        if filter == PhotometricFilter.FILTER_V:
            return self._differential_counts_color(mag, self._null)
        elif filter == PhotometricFilter.FILTER_K:
            return self._differential_counts_color(mag, self._colorsVK)

    def integral_counts(self, mag, filter=PhotometricFilter.FILTER_K):
        if filter == PhotometricFilter.FILTER_K:
            return self._integral_counts_color(mag, self._colorsVK)
        return 0.

    @classmethod
    def _integral_counts_color(cls, mag, colors):
        def subint(i):
            return max(sil.splint(cls._min + colors[i],
                                  mag + colors[i],
                                  cls._spl[i]),
                       0.0
                       )
        return sum(subint(i) for i in (0, 1, 2))

    @classmethod
    def _differential_counts_color(cls, mag, colors):
        def subspl(i):
            return max(sil.splev(mag, cls._spl[i]), 0.0)
        return sum(subspl(i) for i in (0, 1, 2))


class SpagnaModel:
    """Star counts model from Spagna 1999.

    Number of stars per square arc minute
    """

    _J_counts_data = loadtxt(StringIO(
        pkgutil.get_data('emirdrp.simulation', 'spagna-J.dat')
        )
    )
    # Data in file is for square degree
    _J_counts_data[:, 1:3] /= 3600.0
    _spl_J_1 = sil.splrep(_J_counts_data[:, 0], _J_counts_data[:, 1])
    _spl_J_2 = sil.splrep(_J_counts_data[:, 0], _J_counts_data[:, 2])
    del _J_counts_data

    _K_counts_data = loadtxt(StringIO(
        pkgutil.get_data('emirdrp.simulation', 'spagna-K.dat')
        )
    )

    # Data in file is for square degree
    _K_counts_data[:, 1:3] /= 3600.0
    _spl_K_1 = sil.splrep(_K_counts_data[:, 0], _K_counts_data[:, 1])
    _spl_K_2 = sil.splrep(_K_counts_data[:, 0], _K_counts_data[:, 2])
    del _K_counts_data

    name = 'Spagna 1999'
    filters = [PhotometricFilter.FILTER_K, PhotometricFilter.FILTER_J]

    @classmethod
    def integral_counts(cls, mag, filter=PhotometricFilter.FILTER_K):
        if filter == PhotometricFilter.FILTER_J:
            return sil.splev(mag, cls._spl_J_2)
        elif filter == PhotometricFilter.FILTER_K:
            return sil.splev(mag, cls._spl_K_2)
        else:
            return 0.

    @classmethod
    def differential_counts(cls, mag, filter=PhotometricFilter.FILTER_K):
        if filter == PhotometricFilter.FILTER_J:
            return sil.splev(mag, cls._spl_J_1) / 3600.0
        elif filter == PhotometricFilter.FILTER_K:
            return sil.splev(mag, cls._spl_K_1) / 3600.0
        else:
            return 0.


class BSModel:
    """Star counts model from Bahcall & Soneira.

    Number of stars per square arc minute
    """
    name = "Bahcall & Soneira"
    params = array(
        [
            [200, - 0.2, 0.01, 2, 15, 400, - 0.26, 0.065, 1.5, 17.5],
            [925, - 0.132, 0.035, 3, 15.75, 1050, - 0.18, 0.087, 2.5, 17.5],
            [235, - 0.227, 0.0, 1.5, 17, 370, - 0.175, 0.06, 2.0, 18],
            [950, - 0.124, 0.027, 3.1, 16.60, 910, - 0.167, 0.083, 2.5, 18]
        ]
    )

    @classmethod
    def integral_counts(cls, mag, filter=PhotometricFilter.FILTER_V):
        if filter == PhotometricFilter.FILTER_V:
            return cls.dFunction(mag, 0., 0.5 * pi, *cls.params[1])
        elif filter == PhotometricFilter.FILTER_B:
            return cls.dFunction(mag, 0., 0.5 * pi, *cls.params[4])
        else:
            return 0.

    @classmethod
    def differential_counts(cls, mag, filter=PhotometricFilter.FILTER_V):
        if filter == PhotometricFilter.FILTER_V:
            return cls.dFunction(mag, 0., 0.5 * pi, *cls.params[0])
        elif filter == PhotometricFilter.FILTER_B:
            return cls.dFunction(mag, 0., 0.5 * pi, *cls.params[3])
        else:
            return 0.

    @staticmethod
    def mu(magnitude):
        """Compute parameter \mu of Appendix B."""
        if magnitude <= 12.:
            return 0.03
        if magnitude <= 20:
            return 0.0075 * (magnitude - 12.) + 0.03
        # if magnitude > 20:
        return 0.09

    @staticmethod
    def gamma(magnitude):
        if magnitude <= 12.:
            return 0.36
        if magnitude <= 20:
            return 0.04 * (12. - magnitude) + 0.36
        # if magnitude > 20:
        return 0.04

    @staticmethod
    def sigma(galacticLongitude, galacticLatitude):
        return 1.45 - 0.20 * cos(galacticLongitude) * cos(galacticLatitude)

    @classmethod
    def dFunction(cls, mag, glongitude, glatitude, C1, alpha, beta, delta,
                  mStar, C2, kappa, eta, lambdax, mDagger):
        firstTerm = (C1 * pow(10., beta * (mag - mStar)) /
                     pow(1 + pow(10., alpha * (mag - mStar)), delta) /
                     pow(sin(glatitude) *
                         (1 - cls.mu(mag) / tan(glatitude) * cos(glongitude)),
                         3.0 - 5 * cls.gamma(mag))
                     )
        secondTerm = (C2 * pow(10., eta * (mag - mDagger)) /
                      pow(1 + pow(10., kappa * (mag - mDagger)), lambdax) /
                      pow(1 - cos(glatitude) * cos(glongitude),
                          cls.sigma(glongitude, glatitude))
                      )
        return (firstTerm + secondTerm) / 3600.0


if __name__ == '__main__':
    import numpy as np
    import np.random
    from math import sqrt, log

    rbmodel = RBModel()
    # print rbmodel.integral_counts(18)
    # print rbmodel.differential_counts(18)
    sgmodel = SpagnaModel()
    # print sgmodel.integral_counts(18)
    # print sgmodel.differential_counts(18)
    bsmodel = BSModel()
    # print sgmodel.integral_counts(18)
    # print bsmodel.differential_counts(18)

    class GeneralRandom:
        """ Recipe from http://code.activestate.com/recipes/576556/."""
        def __init__(self, x, p, Nrl=1000):
            self.x = x
            self.pdf = p / p.sum()
            self.cdf = self.pdf.cumsum()
            self.inversecdfbins = Nrl
            self.Nrl = Nrl
            y = np.arange(Nrl) / float(Nrl)
            _delta = 1.0 / Nrl
            self.inversecdf = np.zeros(Nrl)
            self.inversecdf[0] = self.x[0]
            cdf_idx = 0
            for n in range(1, self.inversecdfbins):
                while self.cdf[cdf_idx] < y[n] and cdf_idx < Nrl:
                    cdf_idx += 1
                # Seems a linear interpolation
                self.inversecdf[n] = (self.x[cdf_idx - 1] +
                                      (self.x[cdf_idx] -
                                       self.x[cdf_idx - 1]) *
                                      (y[n] - self.cdf[cdf_idx - 1]) /
                                      (self.cdf[cdf_idx] -
                                       self.cdf[cdf_idx - 1])
                                      )
                if cdf_idx >= Nrl:
                    break
            self.delta_inversecdf = np.concatenate(
                (np.diff(self.inversecdf), [0])
                )

        def random(self, N=1000):
            idx_f = np.random.uniform(size=N, high=self.Nrl - 1)
            idx = np.array(idx_f, 'i')
            y = (self.inversecdf[idx] +
                 (idx_f - idx) * self.delta_inversecdf[idx])
            return y

    scmodel = sgmodel
    magmin = 12.0
    magmax = 25.0

    plate_scale = 0.2
    detector_shape = (2048, 2048)
    pixel_area = detector_shape
    detector_area = (plate_scale ** 2 * detector_shape[0] *
                     detector_shape[1]) / 3600.

    nstars = int(round(detector_area * (scmodel.integral_counts(magmax) -
                                        scmodel.integral_counts(magmin))))
    print('nstars', nstars)

    seeing = 1.0

    scale = 2 * sqrt(2 * log(2))

    delta = 0.1
    steps = int(round((magmax - magmin) / delta))
    x = magmin + delta * np.arange(steps)
    p = np.array([scmodel.differential_counts(i) for i in x])

    g = GeneralRandom(x, p)

    np.random.seed(100)
    y = np.random.uniform(high=pixel_area[0], size=nstars)
    x = np.random.uniform(high=pixel_area[1], size=nstars)
    sigmas = [seeing / scale / plate_scale] * nstars
    ints = [100] * nstars
    mag = g.random(N=nstars)



#
# Copyright 2008-2014 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

"""Preprocessing EMIR readout modes"""

from __future__ import division

import six

from astropy.io import fits

from numina.array import ramp_array, fowler_array

from .core import EMIR_READ_MODES

PREPROC_KEY = 'READPROC'
PREPROC_VAL = True


class ReadModeGuessing(object):
    def __init__(self, mode, info=None):
        self.mode = mode
        self.info = info


def image_readmode(hdulist, default=None):
    header = hdulist[0].header
    if 'READMODE' in header:
        p_readmode = header['READMODE'].lower()
        if p_readmode in EMIR_READ_MODES:
            return ReadModeGuessing(mode=p_readmode,
                                    info={'source': 'keyword'}
                                    )
    # Using heuristics
    shape = hdulist[0].shape
    if len(shape) == 2:
        # A 2D image, mode is single
        return ReadModeGuessing(mode='single', info={'source': 'heuristics'})
    else:
        nd = min(shape)
        if nd == 2:
            # A NXNX2 image
            # mide is cds
            return ReadModeGuessing(mode='cds', info={'source': 'heuristics'})
    # Insufficient information
    if default:
        return ReadModeGuessing(mode=default, info={'source': 'default'})
    else:
        return None


def preprocess_single(hdulist):
    return hdulist


def preprocess_cds(hdulist):
    # CDS is Fowler with just one pair of reads
    return preprocess_fowler(hdulist)


def preprocess_fowler(hdulist):
    hdulist[0].header.update(PREPROC_KEY, PREPROC_VAL)
    # We need:
    tint = 0.0  # Integration time (from first read to last read)
    ts = 0.0  # Time between samples
    gain = 1.0  # Detector gain (number)
    ron = 1.0  # Detector RON (number)
    # A master badpixel mask
    cube = hdulist[0].data

    res, var, npix, mask = fowler_array(cube, ti=tint, ts=ts,
                                        gain=gain, ron=ron,
                                        badpixels=None,
                                        dtype='float32',
                                        saturation=55000.0
                                        )
    hdulist[0].data = res
    varhdu = fits.ImageHDU(var)
    varhdu.update_ext_name('VARIANCE')
    hdulist.append(varhdu)
    nmap = fits.ImageHDU(npix)
    nmap.update_ext_name('MAP')
    hdulist.append(nmap)
    nmask_hdu = fits.ImageHDU(mask)
    nmask_hdu.update_ext_name('MASK')
    hdulist.append(nmask_hdu)
    return hdulist


def preprocess_ramp(hdulist):
    hdulist[0].header.update(PREPROC_KEY, PREPROC_VAL)
    cube = hdulist[0].data
    # We need
    ti = 0.0  # Integration time
    gain = 1.0
    ron = 1.0
    rslt = ramp_array(cube, ti, gain=gain,
                      ron=ron, badpixels=None,
                      dtype='float32', saturation=55000.0
                      )
    result, var, npix, mask = rslt

    hdulist[0].data = result
    varhdu = fits.ImageHDU(var)
    varhdu.update_ext_name('VARIANCE')
    hdulist.append(varhdu)

    nmap = fits.ImageHDU(npix)
    nmap.update_ext_name('MAP')
    hdulist.append(nmap)
    nmask = fits.ImageHDU(mask)
    nmask.update_ext_name('MASK')
    hdulist.append(nmask)
    return hdulist


def fits_wrapper(frame):
    if isinstance(frame, six.string_types):
        return fits.open(frame)
    elif isinstance(frame, fits.HDUList):
        return frame
    else:
        raise TypeError


def preprocess(input_, output):
    with fits_wrapper(input_) as hdulist:
        header = hdulist[0].header
        if 'PREPROC' in header:
            # if the image is preprocessed, do nothing
            if input != output:
                hdulist.writeto(output, overwrite=True)
            return
        # determine the READ mode
        guess = image_readmode(hdulist, 'single')
        if guess is None:
            # We have a problem here
            return

        if guess.mode == 'single':
            hduproc = preprocess_single(hdulist)
        elif guess.mode == 'cds':
            hduproc = preprocess_cds(hdulist)
        elif guess.mode == 'fowler':
            hduproc = preprocess_fowler(hdulist)
        elif guess.mode == 'ramp':
            pass
        else:
            hduproc = preprocess_single(hdulist)

        hduproc.writeto(output, overwrite=True)




import astropy.io.fits as fits
import numpy

import numina.core
import numina.exceptions

from ..stare import StareSpectraRecipe


def create_frame(val=0,  keys=None):
    data = numpy.zeros((10, 10)) + val
    hdu = fits.PrimaryHDU(data)
    if keys:
        for k, v in keys.items():
            hdu.header[k] = v

    return fits.HDUList([hdu])


def create_frame_dark():
    data = numpy.zeros((10, 10))
    hdu = fits.PrimaryHDU(data)
    hdu.header['UUID'] = '11111111111111111111111111111111'
    return fits.HDUList([hdu])


def create_frame_flat():
    data = numpy.ones((10, 10))
    hdu = fits.PrimaryHDU(data)
    hdu.header['UUID'] = '2222222222222222222222222222222'
    return fits.HDUList([hdu])


def test_subs():

    expected_data = 2.5 + numpy.zeros((10, 10))
    keys = {'DATE-OBS': '2001-10-01T12:33:43.345', 'TSUTC1': 1000000001.00034, 'TSUTC2': 1000000100.00034}
    frames1 = numina.core.DataFrame(frame=create_frame(val=3, keys=keys))
    keys = {'DATE-OBS': '2001-10-01T12:35:44.345', 'TSUTC1': 1000000101.00034, 'TSUTC2': 1000000200.00034}
    frames2 = numina.core.DataFrame(frame=create_frame(val=2, keys=keys))

    obsresult = numina.core.ObservationResult()
    obsresult.frames = [frames1, frames2]

    recipe = StareSpectraRecipe()
    rinput = recipe.create_input(
        obresult=obsresult,
        master_dark = numina.core.DataFrame(frame=create_frame_dark()),
        master_flat = numina.core.DataFrame(frame=create_frame_flat())
    )
    result = recipe.run(rinput)

    stare_hdul = result.stare.open()
    assert stare_hdul[0].header['TSUTC1'] == 1000000001.00034
    assert stare_hdul[0].header['TSUTC2'] == 1000000200.00034
    assert 'NUM-DK' in stare_hdul[0].header
    assert stare_hdul[0].header['NUM-NCOM'] == 2
    assert numpy.allclose(expected_data, stare_hdul[0].data)



#
# Copyright 2013-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division
#
import logging

import numpy
import numina.processing as  proc


_logger = logging.getLogger(__name__)


class Checker(proc.Corrector):
    """A Node that checks."""
    def __init__(self):
        super(Checker, self).__init__()

    def run(self, img):
        base = img[0].data
        _logger.debug('running checker after flat')
        # Check NaN and Ceros
        mask1 = ~numpy.isfinite(base)
        if numpy.any(mask1):
            _logger.warning('image has %d NaN', mask1.sum())

        return img



#
# Copyright 2015-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Bar characterization using gradients for EMIR"""

from __future__ import division

from numina.core import Requirement, Result, Parameter
from numina.exceptions import RecipeError
import numina.types.array as tarray

import emirdrp.datamodel as datamodel
from emirdrp.processing.bars import slits_to_ds9_reg, find_bars
from emirdrp.core import EMIR_NBARS
from emirdrp.core.recipe import EmirRecipe
from numina.processing.combine import basic_processing_with_combination

import emirdrp.requirements as reqs
import emirdrp.products as prods


class BarDetectionRecipe(EmirRecipe):

    # Recipe Requirements
    #
    obresult = reqs.ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_flat = reqs.MasterIntensityFlatFieldRequirement()
    master_sky = reqs.MasterSkyRequirement()

    bars_nominal_positions = Requirement(prods.NominalPositions,
                                         'Nominal positions of the bars'
                                         )
    median_filter_size = Parameter(5, 'Size of the median box')
    average_box_row_size = Parameter(7, 'Number of rows to average for fine centering (odd)')
    average_box_col_size = Parameter(21, 'Number of columns to extract for fine centering (odd)')
    fit_peak_npoints = Parameter(3, 'Number of points to use for fitting the peak (odd)')

    # Recipe Products
    frame = Result(prods.ProcessedImage)
    # derivative = Result(prods.ProcessedImage)
    slits = Result(tarray.ArrayType)
    positions3 = Result(tarray.ArrayType)
    positions5 = Result(tarray.ArrayType)
    positions7 = Result(tarray.ArrayType)
    positions9 = Result(tarray.ArrayType)
    DTU = Result(tarray.ArrayType)
    ROTANG = Result(float)
    TSUTC1 = Result(float)
    csupos = Result(tarray.ArrayType)
    csusens = Result(tarray.ArrayType)

    def run(self, rinput):
        self.logger.info('starting processing for bars detection')

        flow = self.init_filters(rinput)

        hdulist = basic_processing_with_combination(rinput, flow=flow)

        hdr = hdulist[0].header
        self.set_base_headers(hdr)

        self.save_intermediate_img(hdulist, 'reduced_image.fits')

        try:
            rotang = hdr['ROTANG']
            tsutc1 = hdr['TSUTC1']
            mecs_hdr = datamodel.get_mecs_header(hdulist)
            dtub, dtur = datamodel.get_dtur_from_header(mecs_hdr)
            csupos = datamodel.get_csup_from_header(mecs_hdr)
            if len(csupos) != 2 * EMIR_NBARS:
                raise RecipeError('Number of CSUPOS != 2 * NBARS')
            csusens = datamodel.get_cs_from_header(hdr)

        except KeyError as error:
            self.logger.error(error)
            raise RecipeError(error)

        self.logger.debug('start finding bars')
        allpos, slits = find_bars(hdulist,
                                  rinput.bars_nominal_positions,
                                  csupos,
                                  dtur,
                                  average_box_row_size=rinput.average_box_row_size,
                                  average_box_col_size=rinput.average_box_col_size,
                                  fit_peak_npoints=rinput.fit_peak_npoints,
                                  median_filter_size=rinput.median_filter_size,
                                  logger=self.logger
                                  )

        self.logger.debug('end finding bars')

        if self.intermediate_results:
            with open('ds9.reg', 'w') as ds9reg:
                slits_to_ds9_reg(ds9reg, slits)

        result = self.create_result(frame=hdulist,
                                    slits=slits,
                                    positions9=allpos[9],
                                    positions7=allpos[7],
                                    positions5=allpos[5],
                                    positions3=allpos[3],
                                    DTU=dtub,
                                    ROTANG=rotang,
                                    TSUTC1=tsutc1,
                                    csupos=csupos,
                                    csusens=csusens,
                                    )
        return result



#
# Copyright 2019 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

"""Return islitlet from Y-pixel coordinate in the rectified image"""

from emirdrp.core import EMIR_NBARS
from emirdrp.core import EMIR_NPIXPERSLIT_RECTIFIED


def get_islitlet(ipixel):
    """Return islilet (from 1 to EMIR_NBARS) from Y-pixel coordinate

    Parameters
    ----------
    ipixel : int
        Y-pixel coordinate in the rectified image.

    Returns
    -------
    islitlet : int
        Slitlet number (from 1 to EMIR_NBARS).

    """

    if ipixel < 1:
        raise ValueError('ipixel={} cannot be < 1'.format(ipixel))

    if ipixel > EMIR_NPIXPERSLIT_RECTIFIED * EMIR_NBARS:
        raise ValueError('ipixel={} cannot be > {}'.format(
            ipixel, EMIR_NPIXPERSLIT_RECTIFIED * EMIR_NBARS
        ))

    islitlet = int(ipixel / EMIR_NPIXPERSLIT_RECTIFIED) + 1

    if ipixel % EMIR_NPIXPERSLIT_RECTIFIED == 0:
        islitlet -= 1

    return islitlet



#
# Copyright 2015 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

'''Test the common module'''

import pytest

import numpy
from numpy.testing import assert_allclose

from ..common import normalize_raw


def test_normalize_raw():

    a = numpy.zeros((10, 10))
    a[:5,:5] = -100
    a[5:,:5] = 10000

    b = normalize_raw(a)

    assert b.min() >=0
    assert b.max() <=1

def test_normalize_raw_2():
    tt = numpy.array([-1.0, 655.350, 6553.50, 1e5])
    rtt = numpy.array([0.0, 0.01, 0.1, 1.0])

    ntt = normalize_raw(tt)

    # In this case, min is 0 and max is 1
    assert ntt.min() == 0.0
    assert ntt.max() == 1.0

    assert_allclose(ntt, rtt)




#
# Copyright 2019-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""
Spectroscopy mode, compute low-frequency flatfield
"""

import contextlib
import logging
import numpy as np
from scipy import ndimage
import uuid

import numina.array.combine as combine
from numina.array.display.ximplotxy import ximplotxy
from numina.array.display.pause_debugplot import pause_debugplot
from numina.array.wavecalib.fix_pix_borders import fix_pix_borders
from numina.array.wavecalib.apply_integer_offsets import apply_integer_offsets
from numina.core import Parameter
from numina.core import Result
from numina.frame.utils import copy_img
from numina.processing.combine import combine_imgs

from emirdrp.core.recipe import EmirRecipe
import emirdrp.datamodel
from emirdrp.instrument.csu_configuration import CsuConfiguration
from emirdrp.processing.wavecal.rectwv_coeff_from_mos_library \
    import rectwv_coeff_from_mos_library
from emirdrp.processing.wavecal.rectwv_coeff_to_ds9 import save_four_ds9
from emirdrp.processing.wavecal.rectwv_coeff_to_ds9 \
    import save_spectral_lines_ds9
from emirdrp.processing.wavecal.slitlet2d import Slitlet2D
import emirdrp.products as prods
import emirdrp.requirements as reqs
from emirdrp.tools.nscan_minmax_frontiers import nscan_minmax_frontiers

from emirdrp.core import EMIR_NAXIS1
from emirdrp.core import EMIR_NAXIS2
from emirdrp.core import EMIR_NBARS
from emirdrp.core import EMIR_MINIMUM_SLITLET_WIDTH_MM
from emirdrp.core import EMIR_MAXIMUM_SLITLET_WIDTH_MM


class SpecFlatLowFreq(EmirRecipe):
    """Process continuum exposures of continuum lamp (lamp ON-OFF)

    """

    logger = logging.getLogger(__name__)

    obresult = reqs.ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_flat = reqs.MasterSpectralFlatFieldRequirement()
    rectwv_coeff = reqs.RectWaveCoeffRequirement(optional=True)
    master_rectwv = reqs.MasterRectWaveRequirement(optional=True)

    # note that 'sum' is not allowed as combination method
    method = Parameter(
        'sigmaclip',
        description='Combination method',
        choices=['mean', 'median', 'sigmaclip']
    )
    method_kwargs = Parameter(
        dict(),
        description='Arguments for combination method',
        optional=True
    )
    minimum_slitlet_width_mm = Parameter(
        float(EMIR_MINIMUM_SLITLET_WIDTH_MM),
        description='Minimum width (mm) for a valid slitlet',
        optional=True
    )
    maximum_slitlet_width_mm = Parameter(
        float(EMIR_MAXIMUM_SLITLET_WIDTH_MM),
        description='Maximum width (mm) for a valid slitlet',
        optional=True
    )
    global_integer_offset_x_pix = Parameter(
        0,
        description='Global offset (pixels) in wavelength direction (integer)',
        optional=True
    )
    global_integer_offset_y_pix = Parameter(
        0,
        description='Global offset (pixels) in spatial direction (integer)',
        optional=True
    )
    nwindow_x_median = Parameter(
        301,
        description='Window size to smooth the flatfield in the spectral '
                    'direction',
        optional=True
    )
    nwindow_y_median = Parameter(
        11,
        description='Window size to smooth the flatfield in the spatial '
                    'direction',
        optional=True
    )
    minimum_fraction = Parameter(
        0.01,
        description='Minimum useful flatfielding value (fraction of maximum)',
        optional=True
    )
    minimum_value_in_output = Parameter(
        0.01,
        description='Minimum value allowed in output: pixels below this value'
                    ' are set to 1.0',
        optional=True
    )
    maximum_value_in_output = Parameter(
        10.0,
        description='Maximum value allowed in output: pixels above this value'
                    ' are set to 1.0',
        optional=True
    )
    debugplot = Parameter(
        0,
        description='Debugger parameter',
        optional=True
    )

    reduced_flatlowfreq = Result(prods.MasterSpectralFlat)

    def run(self, rinput):
        self.logger.info('starting generation of flatlowfreq')

        self.logger.info('rectwv_coeff..........................: {}'.format(
            rinput.rectwv_coeff))
        self.logger.info('master_rectwv.........................: {}'.format(
            rinput.master_rectwv))
        self.logger.info('Minimum slitlet width (mm)............: {}'.format(
            rinput.minimum_slitlet_width_mm))
        self.logger.info('Maximum slitlet width (mm)............: {}'.format(
            rinput.maximum_slitlet_width_mm))
        self.logger.info('Global offset X direction (pixels)....: {}'.format(
            rinput.global_integer_offset_x_pix))
        self.logger.info('Global offset Y direction (pixels)....: {}'.format(
            rinput.global_integer_offset_y_pix))
        self.logger.info('nwindow_x_median......................: {}'.format(
            rinput.nwindow_x_median
        ))
        self.logger.info('nwindow_y_median......................: {}'.format(
            rinput.nwindow_y_median
        ))
        self.logger.info('Minimum fraction......................: {}'.format(
            rinput.minimum_fraction))
        self.logger.info('Minimum value in output...............: {}'.format(
            rinput.minimum_value_in_output))
        self.logger.info('Maximum value in output...............: {}'.format(
            rinput.maximum_value_in_output))

        # check rectification and wavelength calibration information
        if rinput.master_rectwv is None and rinput.rectwv_coeff is None:
            raise ValueError('No master_rectwv nor rectwv_coeff data have '
                             'been provided')
        elif rinput.master_rectwv is not None and \
                rinput.rectwv_coeff is not None:
            self.logger.warning('rectwv_coeff will be used instead of '
                                'master_rectwv')
        if rinput.rectwv_coeff is not None and \
                (rinput.global_integer_offset_x_pix != 0 or
                 rinput.global_integer_offset_y_pix != 0):
            raise ValueError('global_integer_offsets cannot be used '
                             'simultaneously with rectwv_coeff')

        # check headers to detect lamp status (on/off)
        list_lampincd = []
        for fname in rinput.obresult.frames:
            with fname.open() as f:
                list_lampincd.append(f[0].header['lampincd'])

        # check number of images
        nimages = len(rinput.obresult.frames)
        n_on = list_lampincd.count(1)
        n_off = list_lampincd.count(0)
        self.logger.info('Number of images with lamp ON.........: {}'.format(
            n_on))
        self.logger.info('Number of images with lamp OFF........: {}'.format(
            n_off))
        self.logger.info('Total number of images................: {}'.format(
            nimages))
        if n_on == 0:
            raise ValueError('Insufficient number of images with lamp ON')
        if n_on + n_off != nimages:
            raise ValueError('Number of images does not match!')

        # check combination method
        if rinput.method_kwargs == {}:
            method_kwargs = None
        else:
            if rinput.method == 'sigmaclip':
                method_kwargs = rinput.method_kwargs
            else:
                raise ValueError('Unexpected method_kwargs={}'.format(
                    rinput.method_kwargs))

        # build object to proceed with bpm, bias, and dark (not flat)
        flow = self.init_filters(rinput)

        # available combination methods
        method = getattr(combine, rinput.method)

        # basic reduction of images with lamp ON or OFF
        lampmode = {0: 'off', 1: 'on'}
        reduced_image_on = None
        reduced_image_off = None
        for imode in lampmode.keys():
            self.logger.info('starting basic reduction of images with'
                             ' lamp {}'.format(lampmode[imode]))
            tmplist = [rinput.obresult.frames[i] for i, lampincd in
                       enumerate(list_lampincd) if lampincd == imode]
            if len(tmplist) > 0:
                with contextlib.ExitStack() as stack:
                    hduls = [stack.enter_context(fname.open()) for fname
                             in tmplist]
                    reduced_image = combine_imgs(
                        hduls,
                        method=method,
                        method_kwargs=method_kwargs,
                        errors=False,
                        prolog=None
                    )
                if imode == 0:
                    reduced_image_off = flow(reduced_image)
                    hdr = reduced_image_off[0].header
                    self.set_base_headers(hdr)
                    self.save_intermediate_img(reduced_image_off,
                                               'reduced_image_off.fits')
                elif imode == 1:
                    reduced_image_on = flow(reduced_image)
                    hdr = reduced_image_on[0].header
                    self.set_base_headers(hdr)
                    self.save_intermediate_img(reduced_image_on,
                                               'reduced_image_on.fits')
                else:
                    raise ValueError('Unexpected imode={}'.format(imode))

        # computation of ON-OFF
        header_on = reduced_image_on[0].header
        data_on = reduced_image_on[0].data.astype('float32')
        if n_off > 0:
            header_off = reduced_image_off[0].header
            data_off = reduced_image_off[0].data.astype('float32')
        else:
            header_off = None
            data_off = np.zeros_like(data_on)
        reduced_data = data_on - data_off

        # update reduced image header
        reduced_image = self.create_reduced_image(
            rinput,
            reduced_data,
            header_on,
            header_off,
            list_lampincd
        )

        # save intermediate image in work directory
        self.save_intermediate_img(reduced_image, 'reduced_image.fits')

        # define rectification and wavelength calibration coefficients
        if rinput.rectwv_coeff is None:
            rectwv_coeff = rectwv_coeff_from_mos_library(
                reduced_image,
                rinput.master_rectwv
            )
            # set global offsets
            rectwv_coeff.global_integer_offset_x_pix = \
                rinput.global_integer_offset_x_pix
            rectwv_coeff.global_integer_offset_y_pix = \
                rinput.global_integer_offset_y_pix
        else:
            rectwv_coeff = rinput.rectwv_coeff
        # save as JSON in work directory
        self.save_structured_as_json(rectwv_coeff, 'rectwv_coeff.json')
        # ds9 region files (to be saved in the work directory)
        if self.intermediate_results:
            save_four_ds9(rectwv_coeff)
            save_spectral_lines_ds9(rectwv_coeff)

        # apply global offsets (to both, the original and the cleaned version)
        image2d = apply_integer_offsets(
            image2d=reduced_data,
            offx=rectwv_coeff.global_integer_offset_x_pix,
            offy=rectwv_coeff.global_integer_offset_y_pix
        )

        # load CSU configuration
        mecs_header = emirdrp.datamodel.get_mecs_header(reduced_image)
        csu_conf = CsuConfiguration.define_from_header(
            mecs_header
        )
        # determine (pseudo) longslits
        dict_longslits = csu_conf.pseudo_longslits()

        # valid slitlet numbers
        list_valid_islitlets = list(range(1, EMIR_NBARS + 1))
        for idel in rectwv_coeff.missing_slitlets:
            self.logger.info('-> Removing slitlet (not defined): ' + str(idel))
            list_valid_islitlets.remove(idel)
        # filter out slitlets with widths outside valid range
        list_outside_valid_width = []
        for islitlet in list_valid_islitlets:
            slitwidth = csu_conf.csu_bar_slit_width(islitlet)
            if (slitwidth < rinput.minimum_slitlet_width_mm) or \
                    (slitwidth > rinput.maximum_slitlet_width_mm):
                list_outside_valid_width.append(islitlet)
                self.logger.info('-> Removing slitlet (width out of range): ' +
                                 str(islitlet))
        if len(list_outside_valid_width) > 0:
            for idel in list_outside_valid_width:
                list_valid_islitlets.remove(idel)

        # initialize rectified image
        image2d_flatfielded = np.zeros((EMIR_NAXIS2, EMIR_NAXIS1))

        # main loop
        grism_name = rectwv_coeff.tags['grism']
        filter_name = rectwv_coeff.tags['filter']
        cout = '0'
        debugplot = rinput.debugplot
        for islitlet in list(range(1, EMIR_NBARS + 1)):
            if islitlet in list_valid_islitlets:
                # define Slitlet2D object
                slt = Slitlet2D(islitlet=islitlet,
                                rectwv_coeff=rectwv_coeff,
                                debugplot=debugplot)
                if abs(slt.debugplot) > 10:
                    print(slt)

                # extract (distorted) slitlet from the initial image
                slitlet2d = slt.extract_slitlet2d(
                    image_2k2k=image2d,
                    subtitle='original image'
                )

                # rectify slitlet
                slitlet2d_rect = slt.rectify(
                    slitlet2d=slitlet2d,
                    resampling=2,
                    subtitle='original (cleaned) rectified'
                )
                naxis2_slitlet2d, naxis1_slitlet2d = slitlet2d_rect.shape

                if naxis1_slitlet2d != EMIR_NAXIS1:
                    print('naxis1_slitlet2d: ', naxis1_slitlet2d)
                    print('EMIR_NAXIS1.....: ', EMIR_NAXIS1)
                    raise ValueError("Unexpected naxis1_slitlet2d")

                # get useful slitlet region (use boundaries)
                spectrail = slt.list_spectrails[0]
                yy0 = slt.corr_yrect_a + \
                      slt.corr_yrect_b * spectrail(slt.x0_reference)
                ii1 = int(yy0 + 0.5) - slt.bb_ns1_orig
                spectrail = slt.list_spectrails[2]
                yy0 = slt.corr_yrect_a + \
                      slt.corr_yrect_b * spectrail(slt.x0_reference)
                ii2 = int(yy0 + 0.5) - slt.bb_ns1_orig

                # median spectrum
                sp_collapsed = np.median(slitlet2d_rect[ii1:(ii2 + 1), :],
                                         axis=0)

                # smooth median spectrum along the spectral direction
                sp_median = ndimage.median_filter(
                    sp_collapsed,
                    5,
                    mode='nearest'
                )

                ymax_spmedian = sp_median.max()
                y_threshold = ymax_spmedian * rinput.minimum_fraction
                lremove = np.where(sp_median < y_threshold)
                sp_median[lremove] = 0.0

                if abs(slt.debugplot) % 10 != 0:
                    xaxis1 = np.arange(1, naxis1_slitlet2d + 1)
                    title = 'Slitlet#' + str(islitlet) + ' (median spectrum)'
                    ax = ximplotxy(xaxis1, sp_collapsed,
                                   title=title,
                                   show=False,
                                   **{'label': 'collapsed spectrum'})
                    ax.plot(xaxis1, sp_median, label='fitted spectrum')
                    ax.plot([1, naxis1_slitlet2d], 2 * [y_threshold],
                            label='threshold')
                    # ax.plot(xknots, yknots, 'o', label='knots')
                    ax.legend()
                    ax.set_ylim(-0.05 * ymax_spmedian, 1.05 * ymax_spmedian)
                    pause_debugplot(slt.debugplot,
                                    pltshow=True, tight_layout=True)

                # generate rectified slitlet region filled with the
                # median spectrum
                slitlet2d_rect_spmedian = np.tile(sp_median,
                                                  (naxis2_slitlet2d, 1))
                if abs(slt.debugplot) % 10 != 0:
                    slt.ximshow_rectified(
                        slitlet2d_rect=slitlet2d_rect_spmedian,
                        subtitle='rectified, filled with median spectrum'
                    )

                # compute smooth surface
                # clipped region
                slitlet2d_rect_clipped = slitlet2d_rect_spmedian.copy()
                slitlet2d_rect_clipped[:(ii1-1), :] = 0.0
                slitlet2d_rect_clipped[(ii2+2):, :] = 0.0
                # unrectified clipped image
                slitlet2d_unrect_clipped = slt.rectify(
                    slitlet2d=slitlet2d_rect_clipped,
                    resampling=2,
                    inverse=True,
                    subtitle='unrectified, filled with median spectrum '
                             '(clipped)'
                )
                # normalize initial slitlet image (avoid division by zero)
                slitlet2d_norm_clipped = np.zeros_like(slitlet2d)
                for j in range(naxis1_slitlet2d):
                    for i in range(naxis2_slitlet2d):
                        den = slitlet2d_unrect_clipped[i, j]
                        if den == 0:
                            slitlet2d_norm_clipped[i, j] = 1.0
                        else:
                            slitlet2d_norm_clipped[i, j] = \
                                slitlet2d[i, j] / den
                # set to 1.0 one additional pixel at each side (since
                # 'den' above is small at the borders and generates wrong
                # bright pixels)
                slitlet2d_norm_clipped = fix_pix_borders(
                    image2d=slitlet2d_norm_clipped,
                    nreplace=1,
                    sought_value=1.0,
                    replacement_value=1.0
                )
                slitlet2d_norm_clipped = slitlet2d_norm_clipped.transpose()
                slitlet2d_norm_clipped = fix_pix_borders(
                    image2d=slitlet2d_norm_clipped,
                    nreplace=1,
                    sought_value=1.0,
                    replacement_value=1.0
                )
                slitlet2d_norm_clipped = slitlet2d_norm_clipped.transpose()
                slitlet2d_norm_smooth = ndimage.median_filter(
                    slitlet2d_norm_clipped,
                    size=(rinput.nwindow_y_median, rinput.nwindow_x_median),
                    mode='nearest'
                )

                if abs(slt.debugplot) % 10 != 0:
                    slt.ximshow_unrectified(
                        slitlet2d=slitlet2d_norm_clipped,
                        subtitle='unrectified, pixel-to-pixel (clipped)'
                    )
                    slt.ximshow_unrectified(
                        slitlet2d=slitlet2d_norm_smooth,
                        subtitle='unrectified, pixel-to-pixel (smoothed)'
                    )

                # ---

                # check for (pseudo) longslit with previous and next slitlet
                imin = dict_longslits[islitlet].imin()
                imax = dict_longslits[islitlet].imax()
                if islitlet > 1:
                    same_slitlet_below = (islitlet - 1) >= imin
                else:
                    same_slitlet_below = False
                if islitlet < EMIR_NBARS:
                    same_slitlet_above = (islitlet + 1) <= imax
                else:
                    same_slitlet_above = False

                for j in range(EMIR_NAXIS1):
                    xchannel = j + 1
                    y0_lower = slt.list_frontiers[0](xchannel)
                    y0_upper = slt.list_frontiers[1](xchannel)
                    n1, n2 = nscan_minmax_frontiers(y0_frontier_lower=y0_lower,
                                                    y0_frontier_upper=y0_upper,
                                                    resize=True)
                    # note that n1 and n2 are scans (ranging from 1 to NAXIS2)
                    nn1 = n1 - slt.bb_ns1_orig + 1
                    nn2 = n2 - slt.bb_ns1_orig + 1
                    image2d_flatfielded[(n1 - 1):n2, j] = \
                        slitlet2d_norm_smooth[(nn1 - 1):nn2, j]

                    # force to 1.0 region around frontiers
                    if not same_slitlet_below:
                        image2d_flatfielded[(n1 - 1):(n1 + 2), j] = 1
                    if not same_slitlet_above:
                        image2d_flatfielded[(n2 - 5):n2, j] = 1
                cout += '.'
            else:
                cout += 'i'

            if islitlet % 10 == 0:
                if cout != 'i':
                    cout = str(islitlet // 10)

            self.logger.info(cout)

        # restore global offsets
        image2d_flatfielded = apply_integer_offsets(
            image2d=image2d_flatfielded,
            offx=-rectwv_coeff.global_integer_offset_x_pix,
            offy=-rectwv_coeff.global_integer_offset_y_pix
        )

        # set pixels below minimum value to 1.0
        filtered = np.where(image2d_flatfielded <
                            rinput.minimum_value_in_output)
        image2d_flatfielded[filtered] = 1.0

        # set pixels above maximum value to 1.0
        filtered = np.where(image2d_flatfielded >
                            rinput.maximum_value_in_output)
        image2d_flatfielded[filtered] = 1.0

        # update image header
        reduced_flatlowfreq = self.create_reduced_image(
            rinput,
            image2d_flatfielded,
            header_on, header_off,
            list_lampincd
        )

        # ds9 region files (to be saved in the work directory)
        if self.intermediate_results:
            save_four_ds9(rectwv_coeff)
            save_spectral_lines_ds9(rectwv_coeff)

        # save results in results directory
        self.logger.info('end of flatlowfreq generation')
        result = self.create_result(
            reduced_flatlowfreq=reduced_flatlowfreq
        )
        return result

    def create_reduced_image(self, rinput, reduced_data,
                             header_on, header_off,
                             list_lampincd):
        with contextlib.ExitStack() as stack:
            hduls = [stack.enter_context(fname.open()) for fname in
                     rinput.obresult.frames]
            # Copy header of first image
            base_header = hduls[0][0].header.copy()
            result = copy_img(hduls[0])
            hdu = result[0]
            hdu.data = reduced_data
            self.set_base_headers(hdu.header)

            self.logger.debug('update result header')
            # update additional keywords
            hdu.header['UUID'] = str(uuid.uuid1())
            hdu.header['OBSMODE'] = 'flatlowfreq'
            hdu.header['TSUTC2'] = hduls[-1][0].header['TSUTC2']
            hdu.header['history'] = "Processed flatlowfreq"
            hdu.header['NUM-NCOM'] = (len(hduls), 'Number of combined frames')

            # update history
            dm = emirdrp.datamodel.EmirDataModel()
            for img, lampincd in zip(hduls, list_lampincd):
                imgid = dm.get_imgid(img)
                hdu.header['HISTORY'] = "Image '{}' has lampincd='{}'".format(
                    imgid, lampincd)
        hdu.header['HISTORY'] = "Processed flatlowfreq"
        hdu.header['HISTORY'] = '--- Reduction of images with lamp ON ---'
        for line in header_on['HISTORY']:
            hdu.header['HISTORY'] = line
        if header_off is not None:
            hdu.header['HISTORY'] = '--- Reduction of images with lamp OFF ---'
            for line in header_off['HISTORY']:
                hdu.header['HISTORY'] = line
        hdu.header.add_history('--- rinput.stored() (BEGIN) ---')
        for item in rinput.stored():
            value = getattr(rinput, item)
            cline = '{}: {}'.format(item, value)
            hdu.header.add_history(cline)
        hdu.header.add_history('--- rinput.stored() (END) ---')
        return result

    def set_base_headers(self, hdr):
        newhdr = super(SpecFlatLowFreq, self).set_base_headers(hdr)
        # Update EXP to 0
        newhdr['EXP'] = 0
        return newhdr



#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division
from __future__ import print_function

import argparse
from datetime import datetime
import numpy as np
import sys
from uuid import uuid4

from numina.array.display.fileinfo import list_fileinfo_from_txt
from numina.array.distortion import ncoef_fmap
from numina.tools.arg_file_is_new import arg_file_is_new
from numina.tools.check_setstate_getstate import check_setstate_getstate
import numina.types.qc

from emirdrp.instrument.dtu_configuration import DtuConfiguration
from emirdrp.products import RefinedBoundaryModelParam
from emirdrp.products import RectWaveCoeff
from emirdrp.products import MasterRectWave

from numina.array.display.pause_debugplot import DEBUGPLOT_CODES
from emirdrp.core import EMIR_NBARS


def islitlet_progress(islitlet, islitlet_max, ignore=False):
    """Auxiliary function to print out progress in loop of slitlets.

    Parameters
    ----------
    islitlet : int
        Current slitlet number.
    islitlet_max : int
        Maximum slitlet number.
    ignore : bool
        If True, display 'i'

    """

    if ignore:
        cout = 'i'
    else:
        if islitlet % 10 == 0:
            cout = str(islitlet // 10)
        else:
            cout = '.'

    sys.stdout.write(cout)
    if islitlet == islitlet_max:
        sys.stdout.write('\n')
    sys.stdout.flush()


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser(prog='rect_wpoly_for_mos')
    # required arguments
    parser.add_argument("input_list",
                        help="TXT file with list JSON files derived from "
                             "longslit data")
    parser.add_argument("--fitted_bound_param", required=True,
                        help="Input JSON with fitted boundary parameters",
                        type=argparse.FileType('rt'))
    parser.add_argument("--out_MOSlibrary", required=True,
                        help="Output JSON file with results",
                        type=lambda x: arg_file_is_new(parser, x))
    # optional arguments
    parser.add_argument("--debugplot",
                        help="Integer indicating plotting & debugging options"
                             " (default=0)",
                        default=0, type=int,
                        choices=DEBUGPLOT_CODES)
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")
    args = parser.parse_args(args)

    if args.echo:
        print('\033[1m\033[31m% ' + ' '.join(sys.argv) + '\033[0m\n')

    # ---

    # Read input TXT file with list of JSON files
    list_json_files = list_fileinfo_from_txt(args.input_list)
    nfiles = len(list_json_files)
    if abs(args.debugplot) >= 10:
        print('>>> Number of input JSON files:', nfiles)
        for item in list_json_files:
            print(item)
    if nfiles < 2:
        raise ValueError("Insufficient number of input JSON files")

    # read fitted boundary parameters and check that all the longslit JSON
    # files have been computed using the same fitted boundary parameters
    refined_boundary_model = RefinedBoundaryModelParam._datatype_load(
        args.fitted_bound_param.name)
    for ifile in range(nfiles):
        coef_rect_wpoly = RectWaveCoeff._datatype_load(
            list_json_files[ifile].filename)
        uuid_tmp = coef_rect_wpoly.meta_info['origin']['bound_param']
        if uuid_tmp[4:] != refined_boundary_model.uuid:
            print('Expected uuid:', refined_boundary_model.uuid)
            print('uuid for ifile #' + str(ifile + 1) + ": " + uuid_tmp)
            raise ValueError("Fitted boundary parameter uuid's do not match")

    # check consistency of grism, filter, DTU configuration and list of
    # valid slitlets
    coef_rect_wpoly_first_longslit = RectWaveCoeff._datatype_load(
        list_json_files[0].filename)
    filter_name = coef_rect_wpoly_first_longslit.tags['filter']
    grism_name = coef_rect_wpoly_first_longslit.tags['grism']
    dtu_conf = DtuConfiguration.define_from_dictionary(
        coef_rect_wpoly_first_longslit.meta_info['dtu_configuration']
    )
    list_valid_islitlets = list(range(1, EMIR_NBARS + 1))
    for idel in coef_rect_wpoly_first_longslit.missing_slitlets:
        list_valid_islitlets.remove(idel)
    for ifile in range(1, nfiles):
        coef_rect_wpoly = RectWaveCoeff._datatype_load(
            list_json_files[ifile].filename)
        filter_tmp = coef_rect_wpoly.tags['filter']
        if filter_name != filter_tmp:
            print(filter_name)
            print(filter_tmp)
            raise ValueError("Unexpected different filter found")
        grism_tmp = coef_rect_wpoly.tags['grism']
        if grism_name != grism_tmp:
            print(grism_name)
            print(grism_tmp)
            raise ValueError("Unexpected different grism found")
        coef_rect_wpoly = RectWaveCoeff._datatype_load(
            list_json_files[ifile].filename)
        dtu_conf_tmp = DtuConfiguration.define_from_dictionary(
            coef_rect_wpoly.meta_info['dtu_configuration']
        )
        if dtu_conf != dtu_conf_tmp:
            print(dtu_conf)
            print(dtu_conf_tmp)
            raise ValueError("Unexpected different DTU configurations found")
        list_valid_islitlets_tmp = list(range(1, EMIR_NBARS + 1))
        for idel in coef_rect_wpoly.missing_slitlets:
            list_valid_islitlets_tmp.remove(idel)
        if list_valid_islitlets != list_valid_islitlets_tmp:
            print(list_valid_islitlets)
            print(list_valid_islitlets_tmp)
            raise ValueError("Unexpected different list of valid slitlets")

    # check consistency of horizontal bounding box limits (bb_nc1_orig and
    # bb_nc2_orig) and ymargin_bb, and store the values for each slitlet
    dict_bb_param = {}
    print("Checking horizontal bounding box limits and ymargin_bb:")
    for islitlet in list(range(1, EMIR_NBARS + 1)):
        if islitlet in list_valid_islitlets:
            islitlet_progress(islitlet, EMIR_NBARS, ignore=False)
            cslitlet = 'slitlet' + str(islitlet).zfill(2)
            dict_bb_param[cslitlet] = {}
            for par in ['bb_nc1_orig', 'bb_nc2_orig', 'ymargin_bb']:
                value_initial = \
                    coef_rect_wpoly_first_longslit.contents[islitlet - 1][par]
                for ifile in range(1, nfiles):
                    coef_rect_wpoly = RectWaveCoeff._datatype_load(
                        list_json_files[ifile].filename)
                    value_tmp = coef_rect_wpoly.contents[islitlet - 1][par]
                    if value_initial != value_tmp:
                        print(islitlet, value_initial, value_tmp)
                        print(value_tmp)
                        raise ValueError("Unexpected different " + par)
                    dict_bb_param[cslitlet][par] = value_initial
        else:
            islitlet_progress(islitlet, EMIR_NBARS, ignore=True)
    print('OK!')

    # ---

    # Read and store all the longslit data
    list_coef_rect_wpoly = []
    for ifile in range(nfiles):
        coef_rect_wpoly = RectWaveCoeff._datatype_load(
            list_json_files[ifile].filename)
        list_coef_rect_wpoly.append(coef_rect_wpoly)

    # ---

    # Initialize structure to save results into an ouptut JSON file
    outdict = {}
    outdict['refined_boundary_model'] = refined_boundary_model.__getstate__()
    outdict['instrument'] = 'EMIR'
    outdict['meta_info'] = {}
    outdict['meta_info']['creation_date'] = datetime.now().isoformat()
    outdict['meta_info']['description'] = \
        'rectification and wavelength calibration polynomial coefficients ' \
        'as a function of csu_bar_slit_center for MOS'
    outdict['meta_info']['recipe_name'] = 'undefined'
    outdict['meta_info']['origin'] = {}
    outdict['meta_info']['origin']['wpoly_longslits'] = {}
    for ifile in range(nfiles):
        cdum = 'longslit_' + str(ifile + 1).zfill(3) + '_uuid'
        outdict['meta_info']['origin']['wpoly_longslits'][cdum] = \
            list_coef_rect_wpoly[ifile].uuid
    outdict['tags'] = {}
    outdict['tags']['grism'] = grism_name
    outdict['tags']['filter'] = filter_name
    outdict['dtu_configuration'] = dtu_conf.outdict()
    outdict['uuid'] = str(uuid4())
    outdict['contents'] = {}

    # include bb_nc1_orig, bb_nc2_orig and ymargin_bb for each slitlet
    # (note that the values of bb_ns1_orig and bb_ns2_orig cannot be
    # computed at this stage because they depend on csu_bar_slit_center)
    for islitlet in list_valid_islitlets:
        cslitlet = 'slitlet' + str(islitlet).zfill(2)
        outdict['contents'][cslitlet] = dict_bb_param[cslitlet]

    # check that order for rectification transformations is the same for all
    # the slitlets and longslit configurations
    order_check_list = []
    for ifile in range(nfiles):
        tmpdict = list_coef_rect_wpoly[ifile].contents
        for islitlet in list_valid_islitlets:
            ttd_order = tmpdict[islitlet - 1]['ttd_order']
            if ttd_order is not None:
                order_check_list.append(ttd_order)
            ttd_order_modeled = \
                tmpdict[islitlet - 1]['ttd_order_longslit_model']
            order_check_list.append(ttd_order_modeled)
    # remove duplicates in list
    order_no_duplicates = list(set(order_check_list))
    if len(order_no_duplicates) != 1:
        print('order_no_duplicates:', order_no_duplicates)
        raise ValueError('tdd_order is not constant!')
    ttd_order = int(order_no_duplicates[0])
    ncoef_rect = ncoef_fmap(ttd_order)
    if abs(args.debugplot) >= 10:
        print('>>> ttd_order........:', ttd_order)
        print('>>> ncoef_rect.......:', ncoef_rect)

    # check that polynomial degree in frontiers and spectrails are the same
    poldeg_check_list = []
    for ifile in range(nfiles):
        tmpdict = list_coef_rect_wpoly[ifile].contents
        for islitlet in list_valid_islitlets:
            tmppoly = tmpdict[islitlet - 1]['frontier']['poly_coef_lower']
            poldeg_check_list.append(len(tmppoly) - 1)
            tmppoly = tmpdict[islitlet - 1]['frontier']['poly_coef_upper']
            poldeg_check_list.append(len(tmppoly) - 1)
            tmppoly = tmpdict[islitlet - 1]['spectrail']['poly_coef_lower']
            poldeg_check_list.append(len(tmppoly) - 1)
            tmppoly = tmpdict[islitlet - 1]['spectrail']['poly_coef_middle']
            poldeg_check_list.append(len(tmppoly) - 1)
            tmppoly = tmpdict[islitlet - 1]['spectrail']['poly_coef_upper']
            poldeg_check_list.append(len(tmppoly) - 1)
    # remove duplicates in list
    poldeg_no_duplicates = list(set(poldeg_check_list))
    if len(poldeg_no_duplicates) != 1:
        print('poldeg_no_duplicates:', poldeg_no_duplicates)
        raise ValueError('poldeg is not constant in frontiers and '
                         'spectrails!')
    poldeg_spectrails = int(poldeg_no_duplicates[0])
    if abs(args.debugplot) >= 10:
        print('>>> poldeg spectrails:', poldeg_spectrails)

    # check that polynomial degree of wavelength calibration is the same for
    # all the slitlets
    poldeg_check_list = []
    for ifile in range(nfiles):
        tmpdict = list_coef_rect_wpoly[ifile].contents
        for islitlet in list_valid_islitlets:
            tmppoly = tmpdict[islitlet - 1]['wpoly_coeff']
            poldeg_check_list.append(len(tmppoly) - 1)
            tmppoly = tmpdict[islitlet - 1]['wpoly_coeff_longslit_model']
            poldeg_check_list.append(len(tmppoly) - 1)
    # remove duplicates in list
    poldeg_no_duplicates = list(set(poldeg_check_list))
    if len(poldeg_no_duplicates) != 1:
        print('poldeg_no_duplicates:', poldeg_no_duplicates)
        raise ValueError('poldeg is not constant in wavelength calibration '
                         'polynomials!')
    poldeg_wavecal = int(poldeg_no_duplicates[0])
    if abs(args.debugplot) >= 10:
        print('>>> poldeg wavecal...:', poldeg_wavecal)

    # ---

    # csu_bar_slit_center values for each slitlet
    print("CSU_bar_slit_center values:")
    for islitlet in list(range(1, EMIR_NBARS + 1)):
        if islitlet in list_valid_islitlets:
            islitlet_progress(islitlet, EMIR_NBARS, ignore=False)
            cslitlet = 'slitlet' + str(islitlet).zfill(2)
            list_csu_bar_slit_center = []
            for ifile in range(nfiles):
                tmpdict = list_coef_rect_wpoly[ifile].contents[islitlet - 1]
                csu_bar_slit_center = tmpdict['csu_bar_slit_center']
                list_csu_bar_slit_center.append(csu_bar_slit_center)
            # check that list_csu_bar_slit_center is properly sorted
            if not np.all(list_csu_bar_slit_center[:-1] <=
                      list_csu_bar_slit_center[1:]):
                print('cslitlet: ', cslitlet)
                print('list_csu_bar_slit_center: ', list_csu_bar_slit_center)
                raise ValueError('Unsorted list_csu_bar_slit_center')
            outdict['contents'][cslitlet]['list_csu_bar_slit_center'] = \
                list_csu_bar_slit_center
        else:
            islitlet_progress(islitlet, EMIR_NBARS, ignore=True)
    print('OK!')

    # ---

    # rectification polynomial coefficients

    # note: when aij and bij have not been computed, we use the modeled
    # version aij_longslit_model and bij_longslit_model
    print("Rectification polynomial coefficients:")
    for islitlet in list(range(1, EMIR_NBARS + 1)):
        if islitlet in list_valid_islitlets:
            islitlet_progress(islitlet, EMIR_NBARS, ignore=False)
            cslitlet = 'slitlet' + str(islitlet).zfill(2)
            outdict['contents'][cslitlet]['ttd_order'] = ttd_order
            outdict['contents'][cslitlet]['ncoef_rect'] = ncoef_rect
            for keycoef in ['ttd_aij', 'ttd_bij', 'tti_aij', 'tti_bij']:
                for icoef in range(ncoef_rect):
                    ccoef = str(icoef).zfill(2)
                    list_cij = []
                    for ifile in range(nfiles):
                        tmpdict = \
                            list_coef_rect_wpoly[ifile].contents[islitlet - 1]
                        cij = tmpdict[keycoef]
                        if cij is not None:
                            list_cij.append(cij[icoef])
                        else:
                            cij_modeled = tmpdict[keycoef + '_longslit_model']
                            if cij_modeled is None:
                                raise ValueError("Unexpected cij_modeled=None!")
                            else:
                                list_cij.append(cij_modeled[icoef])
                            if abs(args.debugplot) >= 10:
                                print("Warning: using " + keycoef +
                                      "_longslit_model for " + cslitlet +
                                      " in file " +
                                      list_json_files[ifile].filename)
                    cdum = 'list_' + keycoef + '_' + ccoef
                    outdict['contents'][cslitlet][cdum] = list_cij
        else:
            islitlet_progress(islitlet, EMIR_NBARS, ignore=True)

    print('OK!')

    # ---

    # wavelength calibration polynomial coefficients

    # note: when wpoly_coeff have not been computed, we use the
    # wpoly_coeff_longslit_model
    print("Wavelength calibration polynomial coefficients:")
    for islitlet in list(range(1, EMIR_NBARS + 1)):
        if islitlet in list_valid_islitlets:
            islitlet_progress(islitlet, EMIR_NBARS, ignore=False)
            cslitlet = 'slitlet' + str(islitlet).zfill(2)
            outdict['contents'][cslitlet]['wpoly_degree'] = poldeg_wavecal
            for icoef in range(poldeg_wavecal + 1):
                ccoef = str(icoef).zfill(2)
                list_cij = []
                for ifile in range(nfiles):
                    tmpdict = list_coef_rect_wpoly[ifile].contents[islitlet - 1]
                    cij = tmpdict['wpoly_coeff']
                    if cij is not None:
                        list_cij.append(cij[icoef])
                    else:
                        cij_modeled = tmpdict['wpoly_coeff_longslit_model']
                        if cij_modeled is None:
                            raise ValueError("Unexpected cij_modeled=None!")
                        else:
                            list_cij.append(cij_modeled[icoef])
                        if abs(args.debugplot) >= 10:
                            print("Warning: using wpoly_coeff_longslit_model" +
                                  " for " + cslitlet +
                                  " in file " +
                                  list_json_files[ifile].filename)
                outdict['contents'][cslitlet]['list_wpoly_coeff_' + ccoef] = \
                    list_cij
        else:
            islitlet_progress(islitlet, EMIR_NBARS, ignore=True)
    print('OK!')

    # ---

    # OBSOLETE
    # Save resulting JSON structure
    '''
    with open(args.out_MOSlibrary.name + '_old', 'w') as fstream:
        json.dump(outdict, fstream, indent=2, sort_keys=True)
        print('>>> Saving file ' + args.out_MOSlibrary.name + '_old')
    '''

    # --

    # Create object of type MasterRectWave with library of coefficients
    # for rectification and wavelength calibration
    master_rectwv = MasterRectWave(instrument='EMIR')
    master_rectwv.quality_control = numina.types.qc.QC.GOOD
    master_rectwv.tags['grism'] = grism_name
    master_rectwv.tags['filter'] = filter_name
    master_rectwv.meta_info['dtu_configuration'] = outdict['dtu_configuration']
    master_rectwv.meta_info['refined_boundary_model'] = {
        'parmodel': refined_boundary_model.meta_info['parmodel']
    }
    master_rectwv.meta_info['refined_boundary_model'].update(
        outdict['refined_boundary_model']['contents']
    )
    master_rectwv.total_slitlets = EMIR_NBARS
    master_rectwv.meta_info['origin'] = {
        'bound_param': 'uuid' + refined_boundary_model.uuid,
        'longslit_frames': ['uuid:' + list_coef_rect_wpoly[ifile].uuid
                            for ifile in range(nfiles)]
    }
    for i in range(EMIR_NBARS):
        islitlet = i + 1
        dumdict = {'islitlet': islitlet}
        cslitlet = 'slitlet' + str(islitlet).zfill(2)
        if cslitlet in outdict['contents']:
            dumdict.update(outdict['contents'][cslitlet])
        else:
            dumdict.update({
                'bb_nc1_orig': 0,
                'bb_nc2_orig': 0,
                'ymargin_bb': 0,
                'list_csu_bar_slit_center': [],
                'ttd_order': 0,
                'ncoef_rect': 0,
                'wpolydegree': 0
            })
            master_rectwv.missing_slitlets.append(islitlet)
        master_rectwv.contents.append(dumdict)
    master_rectwv.writeto(args.out_MOSlibrary.name)
    print('>>> Saving file ' + args.out_MOSlibrary.name)
    # debugging __getstate__ and __setstate__
    # check_setstate_getstate(master_rectwv, args.out_MOSlibrary.name)


if __name__ == "__main__":
    main()



#
# Copyright 2015-2019 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#


"""Test the AIV pinhole mask recipe"""


import pytest

from emirdrp.tests.runrecipe import run_recipe


BASE_URL = 'https://guaix.fis.ucm.es/data/pyemir/test/'


@pytest.mark.remote_data
def _test_mode_TEST6_set0(numinatpldir):
    run_recipe()


@pytest.mark.remote_data
def _test_mode_TEST7_set0(numinatpldir):
    run_recipe()


@pytest.mark.remote_data
def _test_mode_TEST8_set0(numinatpldir):
    run_recipe()



#
# Copyright 2008-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division
from __future__ import print_function

from astropy.io import fits
from copy import deepcopy

from emirdrp.core import EMIR_NBARS
from emirdrp.core import EMIR_MINIMUM_SLITLET_WIDTH_MM
from emirdrp.core import EMIR_MAXIMUM_SLITLET_WIDTH_MM


class LongSlit(object):
    """Auxiliary class to store the minimum and maximum islitlet"""

    def __init__(self, imin=None, imax=None):
        self._imin = imin
        self._imax = imax

    def __repr__(self):
        output = 'islitlet_min_max=({0:02d}, {1:02d})'.format(
            self._imin, self._imax
        )
        return output

    def __eq__(self, other):
        if isinstance(other, LongSlit):
            result = \
                (self._imin == other._imin) and (self._imax == other._imax)
            return result
        return NotImplemented

    def imin(self):
        return self._imin

    def imax(self):
        return self._imax


class CsuConfiguration(object):
    """Configurable Slit Unit (CSU) Configuration class definition.

    Attributes
    ----------
    _csu_bar_left : list of floats
        Location (mm) of the left bar for each slitlet.
    _csu_bar_right : list of floats
        Location (mm) of the right bar for each slitlet, using the
        same origin employed for _csu_bar_left (which is not the
        value stored in the FITS keywords.
    _csu_bar_slit_center : list of floats
        Middle point (mm) in between the two bars defining a slitlet.
    _csu_bar_slit_width : list of floats
        Slitlet width (mm), computed as the distance between the two
        bars defining the slitlet.

    """

    def __init__(self):
        self._csu_bar_left = None
        self._csu_bar_right = None
        self._csu_bar_slit_center = None
        self._csu_bar_slit_width = None

    def __str__(self):
        output = "<CsuConfiguration instance>\n"
        for i in range(EMIR_NBARS):
            ibar = i + 1
            strdum = "- [BAR{0:2d}] left, right, center, width: ".format(ibar)
            output += strdum
            strdum = "{0:7.3f} {1:7.3f} {2:7.3f} {3:7.3f}\n".format(
                self._csu_bar_left[i], self._csu_bar_right[i],
                self._csu_bar_slit_center[i], self._csu_bar_slit_width[i]
            )
            output += strdum
        return output

    def __eq__(self, other):
        if isinstance(other, CsuConfiguration):
            ndig = 3
            result = \
                (round(self._csu_bar_left, ndig) ==
                 round(other._csu_bar_left, ndig)) and \
                (round(self._csu_bar_right, ndig) ==
                 round(other._csu_bar_right, ndig)) and \
                (round(self._csu_bar_slit_center) ==
                 round(other._csu_bar_slit_center)) and \
                (round(self._csu_bar_slit_width, ndig) ==
                 round(other._csu_bar_slit_width, ndig))
            return result
        return NotImplemented

    @classmethod
    def define_from_fits(cls, fitsobj, extnum=None):
        """Define class members from header information in FITS file.

        Parameters
        ----------
        fitsobj: file object
            FITS file whose header contains the CSU bar information
            needed to initialise the members of this class.
        extnum : None or int or str
            Extension number or extension name. If None, select 'MECS' extension
            if available. If not, use priamry extension

        """

        # read input FITS file
        with fits.open(fitsobj) as hdulist:
            if extnum is None:
                if 'MECS' in hdulist:
                    extnum = 'MECS'
                else:
                    extnum = 0
            image_header = hdulist[extnum].header
            return cls.define_from_header(image_header)

    @classmethod
    def define_from_header(cls, image_header):
        """Define class members directly from FITS header.

        Parameters
        ----------
        image_header : instance of hdulist.header
            Header content from a FITS file.

        """

        self = CsuConfiguration()

        # declare lists to store configuration of CSU bars
        self._csu_bar_left = []
        self._csu_bar_right = []
        self._csu_bar_slit_center = []
        self._csu_bar_slit_width = []

        for i in range(EMIR_NBARS):
            ibar = i + 1
            keyword = 'CSUP{}'.format(ibar)
            if keyword in image_header:
                self._csu_bar_left.append(image_header[keyword])
            else:
                raise ValueError("Expected keyword " + keyword + " not found!")
            keyword = 'CSUP{}'.format(ibar + EMIR_NBARS)
            if keyword in image_header:
                # set the same origin as the one employed for _csu_bar_left
                self._csu_bar_right.append(341.5 - image_header[keyword])
            else:
                raise ValueError("Expected keyword " + keyword + " not found!")
            self._csu_bar_slit_center.append(
                (self._csu_bar_left[i] + self._csu_bar_right[i]) / 2
            )
            self._csu_bar_slit_width.append(
                self._csu_bar_right[i] - self._csu_bar_left[i]
            )

        return self

    def csu_bar_left(self, islitlet):
        """Return csu_bar_left for requested slitlet number."""

        return self._csu_bar_left[islitlet - 1]

    def csu_bar_right(self, islitlet):
        """Return csu_bar_right for requested slitlet number."""

        return self._csu_bar_right[islitlet - 1]

    def csu_bar_slit_center(self, islitlet):
        """Return csu_bar_slit_center for requested slitlet number."""

        return self._csu_bar_slit_center[islitlet - 1]

    def csu_bar_slit_width(self, islitlet):
        """Return csu_bar_slit_width for requested slitlet number."""

        return self._csu_bar_slit_width[islitlet - 1]

    def outdict(self, ndigits=3):
        """Return dictionary structure rounded to a given precision."""

        outdict = {}
        for i in range(EMIR_NBARS):
            ibar = i + 1
            cbar = 'slitlet' + str(ibar).zfill(2)
            outdict[cbar] = {}
            outdict[cbar]['_csu_bar_left'] = \
                round(self._csu_bar_left[i], ndigits)
            outdict[cbar]['_csu_bar_right'] = \
                round(self._csu_bar_right[i], ndigits)
            outdict[cbar]['_csu_bar_slit_center'] = \
                round(self._csu_bar_slit_center[i], ndigits)
            outdict[cbar]['_csu_bar_slit_width'] = \
                round(self._csu_bar_slit_width[i], ndigits)

        return outdict

    def widths_in_range_mm(
            self,
            minwidth=EMIR_MINIMUM_SLITLET_WIDTH_MM,
            maxwidth=EMIR_MAXIMUM_SLITLET_WIDTH_MM
    ):
        """Return list of slitlets which width is within given range

        Parameters
        ----------
        minwidth : float
            Minimum slit width (mm).
        maxwidth : float
            Maximum slit width (mm).

        Returns
        -------
        list_ok : list
            List of booleans indicating whether the corresponding
            slitlet width is within range

        """

        list_ok = []
        for i in range(EMIR_NBARS):
            slitlet_ok = minwidth <= self._csu_bar_slit_width[i] <= maxwidth
            if slitlet_ok:
                list_ok.append(i + 1)

        return list_ok

    def pseudo_longslits(self, list_valid_slitlets=None,
                         fracwidth=0.25, diffcenter=0.25):
        """Return dictionary of LongSlit instances

        Parameters
        ----------
        list_valid_slitlets : list of integers or 'all'
            List containing valid slitlets. Slitlet numbers not
            included in this list will be considered as individual
            longslits. If this list is None, all the slitlets are
            analysed.
        fracwidth : float
            Maximum fractional slitlet width difference allowed
            to consider that two slitlets belong to the same longslit.
        diffcenter : float
            Maximum allowed difference between slitlet centers (in
            units of the averaged slitlet widths).

        Returns
        -------
        result : dictionary of LongSlit instances
            Each LongSlit object provides the minimum and maximum
            islitlet for each longslit.

        """

        if list_valid_slitlets is None:
            list_valid_slitlets = range(1, EMIR_NBARS + 1)

        result = dict()
        for islitlet in range(1, EMIR_NBARS + 1):
            result[islitlet] = LongSlit(islitlet, islitlet)

        # check for pseudo-longslit with previous slitlet
        for islitlet in range(2, EMIR_NBARS + 1):
            if islitlet in list_valid_slitlets:
                if (islitlet - 1) in list_valid_slitlets:
                    c1 = self.csu_bar_slit_center(islitlet - 1)
                    w1 = self.csu_bar_slit_width(islitlet - 1)
                    c2 = self.csu_bar_slit_center(islitlet)
                    w2 = self.csu_bar_slit_width(islitlet)
                    wmean = (w1 + w2) / 2.0
                    if abs(w1 - w2) / wmean < fracwidth:
                        if abs(c1 - c2) < wmean * diffcenter:
                            result[islitlet]._imin = \
                                result[(islitlet - 1)]._imin

        # check for pseudo-longslit with next slitlet
        for islitlet in reversed(range(1, EMIR_NBARS)):
            if islitlet in list_valid_slitlets:
                if (islitlet + 1) in list_valid_slitlets:
                    c1 = self.csu_bar_slit_center(islitlet)
                    w1 = self.csu_bar_slit_width(islitlet)
                    c2 = self.csu_bar_slit_center(islitlet + 1)
                    w2 = self.csu_bar_slit_width(islitlet + 1)
                    wmean = (w1 + w2) / 2.0
                    if abs(w1 - w2) / wmean < fracwidth:
                        if abs(c1 - c2) < wmean * diffcenter:
                            result[islitlet]._imax = \
                                result[(islitlet + 1)]._imax

        return result

    def display_pseudo_longslits(self, list_valid_slitlets=None,
                                 fracwidth=0.25, diffcenter=0.25):
        dict_longslits = self.pseudo_longslits(
            list_valid_slitlets=list_valid_slitlets,
            fracwidth=fracwidth,
            diffcenter=diffcenter
        )
        for islitlet in range(1, EMIR_NBARS + 1):
            longslit = dict_longslits[islitlet]
            imin = longslit._imin
            imax = longslit._imax
            output = 'islitlet: {0:02d} '.format(islitlet)
            output += '--> min: {0:02d}  max: {1:02d} '.format(imin, imax)
            if imin == islitlet == imax:
                output += '<->'
            elif imin == islitlet < imax:
                output += '<--'
            elif imin < islitlet == imax:
                output += '-->'
            else:
                output += '...'
            print(output)


def merge_odd_even_csu_configurations(conf_odd, conf_even):
    """Merge CSU configuration using odd- and even-numbered values.

    The CSU returned CSU configuration include the odd-numbered values
    from 'conf_odd' and the even-numbered values from 'conf_even'.

    Parameters
    ----------
    conf_odd : CsuConfiguration instance
        CSU configuration corresponding to odd-numbered slitlets.
    conf_even : CsuConfiguration instance
        CSU configuration corresponding to even-numbered slitlets.

    Returns
    -------
    merged_conf : CsuConfiguration instance
        CSU configuration resulting from the merging process.

    """

    # initialize resulting CsuConfiguration instance using one of the
    # input configuration corresponding to the odd-numbered slitlets
    merged_conf = deepcopy(conf_odd)

    # update the resulting configuration with the values corresponding
    # to the even-numbered slitlets
    for i in range(EMIR_NBARS):
        ibar = i + 1
        if ibar % 2 == 0:
            merged_conf._csu_bar_left[i] = conf_even._csu_bar_left[i]
            merged_conf._csu_bar_right[i] = conf_even._csu_bar_right[i]
            merged_conf._csu_bar_slit_center[i] = \
                conf_even._csu_bar_slit_center[i]
            merged_conf._csu_bar_slit_width[i] = \
                conf_even._csu_bar_slit_width[i]

    # return merged configuration
    return merged_conf



#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

"""Decorator for EMIR Recipes"""

from __future__ import print_function

from numina.core import DataFrame, ObservationResult
import emirdrp.processing.info as info


def loginfo(method):
    """Log the contents of Recipe Input"""


    def loginfo_method(self, rinput):

        klass = rinput.__class__

        for key in klass.stored():
            val = getattr(rinput, key)
            if isinstance(val, DataFrame):
                self.logger.debug("DataFrame %s", info.gather_info_dframe(val))
            elif isinstance(val, ObservationResult):
                for f in val.images:
                    self.logger.debug("OB DataFrame %s" , info.gather_info_dframe(f))
            else:
                pass

        result = method(self, rinput)

        return result

    return loginfo_method


def aggregate(method):

    def agregate_method(self, rinput):

        result = method(self, rinput)

        aggregate_m = getattr(self, 'aggregate_result')

        new_result = aggregate_m(self, result, rinput)

        return new_result

    return agregate_method



#
# Copyright 2008-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import numpy
import astropy.units as U

import emirdrp.instrument.constants as cons

EMIR_PLATESCALE_RADS = cons.EMIR_PIXSCALE.to(U.rad / U.pixel).value


def adapt_wcs(wcs, rotang_deg, ipa_deg):
    """Adapt WCS to use in distortions"""
    wcsa = wcs.deepcopy()
    wcsa.wcs.crval = [0, 0]
    cdelt = numpy.sqrt(numpy.linalg.det(wcs.wcs.cd))
    angle = numpy.arctan2(wcs.wcs.cd[1, 0], wcs.wcs.cd[0, 0])
    nangle = angle + (ipa_deg - rotang_deg) / 180.0 * numpy.pi
    pc11 = cdelt * numpy.cos(nangle)
    pc22 = cdelt * numpy.cos(nangle)
    pc12 = -cdelt * numpy.sin(nangle)
    pc21 = cdelt * numpy.sin(nangle)
    wcsa.wcs.cd = [[pc11, pc12],
                   [pc21, pc22]]
    return wcsa


def exvp(pos_x, pos_y):
    """Convert virtual pixel to real pixel

    This function is deprecated, use wcs_exvp instead
    """
    pos_x = numpy.asarray(pos_x)
    pos_y = numpy.asarray(pos_y)
    # convert virtual pixel to real pixel
    # convert world coordinate to pixel
    center = [1024.5, 1024.5]
    cf = EMIR_PLATESCALE_RADS
    pos_base_x = pos_x - center[0]
    pos_base_y = pos_y - center[1]
    ra = numpy.hypot(pos_base_x, pos_base_y)
    thet = numpy.arctan2(pos_base_y, pos_base_x)

    r  = cf * ra

    rr1 = 1 + 14606.7 * r**2 + 1739716115.1 * r**4

    nx1 = rr1 * ra * numpy.cos(thet) + center[0]
    ny1 = rr1 * ra * numpy.sin(thet) + center[1]
    return nx1, ny1


def wcs_exvp(wcs, xv, yv):
    """Convert virtual pixel to real pixel"""

    # These factors are equivalent to 1 / CDELT1 1 / CDELT2
    cdelt1 = numpy.hypot(wcs.wcs.cd[0, 0], wcs.wcs.cd[0, 1])
    cdelt2 = numpy.hypot(wcs.wcs.cd[1, 0], wcs.wcs.cd[1, 1])
    #
    ra_i = (wcs.wcs.crpix[0] - xv) * cdelt1
    dec_i = (wcs.wcs.crpix[1] - yv) * cdelt2
    # keep old value of crval
    oldcrval = wcs.wcs.crval.copy()
    wcs.wcs.crval = [0, 0]
    xr, yr = wcs.all_world2pix(ra_i, dec_i, 1)
    # reset crval
    wcs.wcs.crval = oldcrval
    return xr, yr


def pvex(pos_x, pos_y):
    """Convert real pixel to virtual pixel


    This function is deprecated, use wcs_pvex instead
    """
    pos_x = numpy.asarray(pos_x)
    pos_y = numpy.asarray(pos_y)

    center = [1024.5, 1024.5]
    cf = EMIR_PLATESCALE_RADS
    pos_base_x = pos_x - center[0]
    pos_base_y = pos_y - center[1]
    ra = numpy.hypot(pos_base_x, pos_base_y)
    thet = numpy.arctan2(pos_base_y, pos_base_x)

    r  = cf * ra
    rr1 = 1.000051 - 14892 * r**2 - 696254464 * r**4

    nx1 = rr1 * ra * numpy.cos(thet) + center[0]
    ny1 = rr1 * ra * numpy.sin(thet) + center[1]
    return nx1, ny1


def wcs_pvex(wcs, xr, yr):
    """Convert real pixel to virtual pixel"""

    # These factors are equivalent to 1 / CDELT1 1 / CDELT2
    cdelt1 = numpy.hypot(wcs.wcs.cd[0, 0], wcs.wcs.cd[0, 1])
    cdelt2 = numpy.hypot(wcs.wcs.cd[1, 0], wcs.wcs.cd[1, 1])
    cfax = 1 / cdelt1
    cfay = 1 / cdelt2

    # keep old value of crval
    oldcrval = wcs.wcs.crval.copy()
    wcs.wcs.crval = [0, 0]
    ra_i, dec_i = wcs.all_pix2world(xr, yr, 1)
    # wrap angles around 180
    # [-180, 180)
    ra_i = numpy.mod(ra_i - 180, 360.0) - 180

    xv = wcs.wcs.crpix[0] - ra_i * cfax
    yv = wcs.wcs.crpix[1] - dec_i * cfay
    # restore old value of crval
    wcs.wcs.crval = oldcrval
    return xv, yv


def pix2virt(pos, origin=1):
    """

    This function is deprecated, use wcs_pix2virt instead
    """
    ddef_o = 1
    off = ddef_o - origin
    pos = numpy.atleast_2d(pos) + off
    nx, ny = pvex(pos[:, 0], pos[:, 1])
    res = numpy.stack((nx, ny), axis=1)
    return res - off


def wcs_pix2virt(wcs, pos, origin=1):
    ddef_o = 1
    off = ddef_o - origin
    pos = numpy.atleast_2d(pos) + off
    nx, ny = wcs_pvex(wcs, pos[:, 0], pos[:, 1])
    res = numpy.stack((nx, ny), axis=1)
    return res - off



from numina.array import combine
import numpy

from ..combine import combine_images, scale_with_median


def create_images():
    import astropy.io.fits as fits
    aa = numpy.array([[1, 2, 3], [6, 5, 4]], dtype='float32')
    bb = numpy.array([[1]], dtype='uint16')
    uu = fits.PrimaryHDU(aa)
    mecs = fits.ImageHDU(bb, name='MECS')
    uu.header['TSUTC2'] = 0
    images = [fits.HDUList([uu, mecs]), fits.HDUList([uu, mecs])]
    return images


def test_combine_images():
    images = create_images()
    res = combine_images(images)
    assert 'MECS' in res


def test_combine_images2():
    images = create_images()
    res = combine_images(images, errors=True)
    assert 'MECS' in res
    assert 'VARIANCE' in res
    assert 'MAP' in res


def test_combine_images3():
    images = create_images()
    scaled_mean = scale_with_median(combine.mean)
    res = combine_images(images, method=scaled_mean, errors=True)
    assert 'MECS' in res
    assert 'VARIANCE' in res
    assert 'MAP' in res



#
# Copyright 2018-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Rectification and wavelength calibration polynomials from empirical library
"""

from __future__ import division, print_function

import argparse
from astropy.io import fits
from datetime import datetime
import logging
import numpy as np
from scipy.interpolate import interp1d
import sys
from uuid import uuid4

from numina.array.distortion import ncoef_fmap
from numina.tools.arg_file_is_new import arg_file_is_new
import numina.types.qc

import emirdrp.datamodel as datamodel
from emirdrp.instrument.csu_configuration import CsuConfiguration
from emirdrp.instrument.dtu_configuration import DtuConfiguration
from emirdrp.products import MasterRectWave
from emirdrp.products import RectWaveCoeff
from emirdrp.tools.fit_boundaries import bound_params_from_dict
from emirdrp.tools.fit_boundaries import expected_distorted_boundaries
from emirdrp.tools.fit_boundaries import expected_distorted_frontiers
from emirdrp.processing.wavecal.slitlet2darc import expected_y0_lower_frontier
from emirdrp.processing.wavecal.slitlet2darc import expected_y0_upper_frontier

from numina.array.display.pause_debugplot import DEBUGPLOT_CODES
from emirdrp.core import EMIR_NAXIS1
from emirdrp.core import EMIR_NAXIS2
from emirdrp.core import EMIR_NBARS


def rectwv_coeff_from_mos_library(reduced_image,
                                  master_rectwv,
                                  ignore_dtu_configuration=True,
                                  debugplot=0):
    """Evaluate rect.+wavecal. coefficients from MOS library

    Parameters
    ----------
    reduced_image : HDUList object
        Image with preliminary basic reduction: bpm, bias, dark and
        flatfield.
    master_rectwv : MasterRectWave instance
        Rectification and Wavelength Calibrartion Library product.
        Contains the library of polynomial coefficients necessary
        to generate an instance of RectWaveCoeff with the rectification
        and wavelength calibration coefficients for the particular
        CSU configuration.
    ignore_dtu_configuration : bool
        If True, ignore differences in DTU configuration.
    debugplot : int
        Debugging level for messages and plots. For details see
        'numina.array.display.pause_debugplot.py'.

    Returns
    -------
    rectwv_coeff : RectWaveCoeff instance
        Rectification and wavelength calibration coefficients for the
        particular CSU configuration.

    """

    logger = logging.getLogger(__name__)
    logger.info('Computing expected RectWaveCoeff from CSU configuration')

    # header
    header = reduced_image[0].header
    mecs_header = datamodel.get_mecs_header(reduced_image)

    # read the CSU configuration from the image header
    csu_conf = CsuConfiguration.define_from_header(mecs_header)

    # read the DTU configuration from the image header
    dtu_conf = DtuConfiguration.define_from_header(mecs_header)

    # retrieve DTU configuration from MasterRectWave object
    dtu_conf_calib = DtuConfiguration.define_from_dictionary(
        master_rectwv.meta_info['dtu_configuration']
    )
    # check that the DTU configuration employed to obtain the calibration
    # corresponds to the DTU configuration in the input FITS file
    if dtu_conf != dtu_conf_calib:
        if ignore_dtu_configuration:
            logger.warning('DTU configuration differences found!')
        else:
            logger.info('DTU configuration from image header:')
            logger.info(dtu_conf)
            logger.info('DTU configuration from master calibration:')
            logger.info(dtu_conf_calib)
            raise ValueError("DTU configurations do not match!")
    else:
        logger.info('DTU configuration match!')

    # check grism and filter
    filter_name = header['filter']
    logger.debug('Filter: ' + filter_name)
    if filter_name != master_rectwv.tags['filter']:
        raise ValueError('Filter name does not match!')
    grism_name = header['grism']
    logger.debug('Grism: ' + grism_name)
    if grism_name != master_rectwv.tags['grism']:
        raise ValueError('Grism name does not match!')

    # valid slitlet numbers
    list_valid_islitlets = list(range(1, EMIR_NBARS + 1))
    for idel in master_rectwv.missing_slitlets:
        list_valid_islitlets.remove(idel)
    logger.debug('valid slitlet numbers: ' + str(list_valid_islitlets))

    # initialize intermediate dictionary with relevant information
    # (note: this dictionary corresponds to an old structure employed to
    # store the information in a JSON file; this is no longer necessary,
    # but here we reuse that dictionary for convenience)
    outdict = {}
    outdict['instrument'] = 'EMIR'
    outdict['meta_info'] = {}
    outdict['meta_info']['creation_date'] = datetime.now().isoformat()
    outdict['meta_info']['description'] = \
        'computation of rectification and wavelength calibration polynomial ' \
        'coefficients for a particular CSU configuration from a MOS model '
    outdict['meta_info']['recipe_name'] = 'undefined'
    outdict['meta_info']['origin'] = {}
    outdict['meta_info']['origin']['fits_frame_uuid'] = 'TBD'
    outdict['meta_info']['origin']['rect_wpoly_mos_uuid'] = \
        master_rectwv.uuid
    outdict['meta_info']['origin']['fitted_boundary_param_uuid'] = \
        master_rectwv.meta_info['origin']['bound_param']
    outdict['tags'] = {}
    outdict['tags']['grism'] = grism_name
    outdict['tags']['filter'] = filter_name
    outdict['dtu_configuration'] = dtu_conf.outdict()
    outdict['uuid'] = str(uuid4())
    outdict['contents'] = {}

    # compute rectification and wavelength calibration coefficients for each
    # slitlet according to its csu_bar_slit_center value
    for islitlet in list_valid_islitlets:
        cslitlet = 'slitlet' + str(islitlet).zfill(2)

        # csu_bar_slit_center of current slitlet in initial FITS image
        csu_bar_slit_center = csu_conf.csu_bar_slit_center(islitlet)

        # input data structure
        tmpdict = master_rectwv.contents[islitlet - 1]
        list_csu_bar_slit_center = tmpdict['list_csu_bar_slit_center']

        # check extrapolations
        if csu_bar_slit_center < min(list_csu_bar_slit_center):
            logger.warning('extrapolating table with ' + cslitlet)
            logger.warning('minimum tabulated value: ' +
                           str(min(list_csu_bar_slit_center)))
            logger.warning('sought value...........: ' +
                           str(csu_bar_slit_center))
        if csu_bar_slit_center > max(list_csu_bar_slit_center):
            logger.warning('extrapolating table with ' + cslitlet)
            logger.warning('maximum tabulated value: ' +
                           str(max(list_csu_bar_slit_center)))
            logger.warning('sought value...........: ' +
                           str(csu_bar_slit_center))

        # rectification coefficients
        ttd_order = tmpdict['ttd_order']
        ncoef = ncoef_fmap(ttd_order)
        outdict['contents'][cslitlet] = {}
        outdict['contents'][cslitlet]['ttd_order'] = ttd_order
        outdict['contents'][cslitlet]['ttd_order_longslit_model'] = None
        for keycoef in ['ttd_aij', 'ttd_bij', 'tti_aij', 'tti_bij']:
            coef_out = []
            for icoef in range(ncoef):
                ccoef = str(icoef).zfill(2)
                list_cij = tmpdict['list_' + keycoef + '_' + ccoef]
                funinterp_coef = interp1d(list_csu_bar_slit_center,
                                          list_cij,
                                          kind='linear',
                                          fill_value='extrapolate')
                # note: funinterp_coef expects a numpy array
                dum = funinterp_coef([csu_bar_slit_center])
                coef_out.append(dum[0])
            outdict['contents'][cslitlet][keycoef] = coef_out
            outdict['contents'][cslitlet][keycoef + '_longslit_model'] = None

        # wavelength calibration coefficients
        ncoef = tmpdict['wpoly_degree'] + 1
        wpoly_coeff = []
        for icoef in range(ncoef):
            ccoef = str(icoef).zfill(2)
            list_cij = tmpdict['list_wpoly_coeff_' + ccoef]
            funinterp_coef = interp1d(list_csu_bar_slit_center,
                                      list_cij,
                                      kind='linear',
                                      fill_value='extrapolate')
            # note: funinterp_coef expects a numpy array
            dum = funinterp_coef([csu_bar_slit_center])
            wpoly_coeff.append(dum[0])
        outdict['contents'][cslitlet]['wpoly_coeff'] = wpoly_coeff
        outdict['contents'][cslitlet]['wpoly_coeff_longslit_model'] = None

        # update cdelt1_linear and crval1_linear
        wpoly_function = np.polynomial.Polynomial(wpoly_coeff)
        crmin1_linear = wpoly_function(1)
        crmax1_linear = wpoly_function(EMIR_NAXIS1)
        cdelt1_linear = (crmax1_linear - crmin1_linear) / (EMIR_NAXIS1 - 1)
        crval1_linear = crmin1_linear
        outdict['contents'][cslitlet]['crval1_linear'] = crval1_linear
        outdict['contents'][cslitlet]['cdelt1_linear'] = cdelt1_linear

        # update CSU keywords
        outdict['contents'][cslitlet]['csu_bar_left'] = \
            csu_conf.csu_bar_left(islitlet)
        outdict['contents'][cslitlet]['csu_bar_right'] = \
            csu_conf.csu_bar_right(islitlet)
        outdict['contents'][cslitlet]['csu_bar_slit_center'] = \
            csu_conf.csu_bar_slit_center(islitlet)
        outdict['contents'][cslitlet]['csu_bar_slit_width'] = \
            csu_conf.csu_bar_slit_width(islitlet)

    # for each slitlet compute spectrum trails and frontiers using the
    # fitted boundary parameters
    fitted_bound_param_json = {
        'contents': master_rectwv.meta_info['refined_boundary_model']
    }
    parmodel = fitted_bound_param_json['contents']['parmodel']
    fitted_bound_param_json.update({'meta_info': {'parmodel': parmodel}})
    params = bound_params_from_dict(fitted_bound_param_json)
    if abs(debugplot) >= 10:
        logger.debug('Fitted boundary parameters:')
        logger.debug(params.pretty_print())
    for islitlet in list_valid_islitlets:
        cslitlet = 'slitlet' + str(islitlet).zfill(2)
        # csu_bar_slit_center of current slitlet in initial FITS image
        csu_bar_slit_center = csu_conf.csu_bar_slit_center(islitlet)
        # compute and store x0_reference value
        x0_reference = float(EMIR_NAXIS1) / 2.0 + 0.5
        outdict['contents'][cslitlet]['x0_reference'] = x0_reference
        # compute spectrum trails (lower, middle and upper)
        list_spectrails = expected_distorted_boundaries(
            islitlet, csu_bar_slit_center,
            [0, 0.5, 1], params, parmodel,
            numpts=101, deg=5, debugplot=0
        )
        # store spectrails in output JSON file
        outdict['contents'][cslitlet]['spectrail'] = {}
        for idum, cdum in zip(range(3), ['lower', 'middle', 'upper']):
            outdict['contents'][cslitlet]['spectrail']['poly_coef_' + cdum] = \
                list_spectrails[idum].poly_funct.coef.tolist()
            outdict['contents'][cslitlet]['y0_reference_' + cdum] = \
                list_spectrails[idum].poly_funct(x0_reference)
        # compute frontiers (lower, upper)
        list_frontiers = expected_distorted_frontiers(
            islitlet, csu_bar_slit_center,
            params, parmodel,
            numpts=101, deg=5, debugplot=0
        )
        # store frontiers in output JSON
        outdict['contents'][cslitlet]['frontier'] = {}
        for idum, cdum in zip(range(2), ['lower', 'upper']):
            outdict['contents'][cslitlet]['frontier']['poly_coef_' + cdum] = \
                list_frontiers[idum].poly_funct.coef.tolist()
            outdict['contents'][cslitlet]['y0_frontier_' + cdum] = \
                list_frontiers[idum].poly_funct(x0_reference)

    # store bounding box parameters for each slitlet
    xdum = np.linspace(1, EMIR_NAXIS1, num=EMIR_NAXIS1)
    for islitlet in list_valid_islitlets:
        cslitlet = 'slitlet' + str(islitlet).zfill(2)
        # parameters already available in the input JSON file
        for par in ['bb_nc1_orig', 'bb_nc2_orig', 'ymargin_bb']:
            outdict['contents'][cslitlet][par] = \
                master_rectwv.contents[islitlet - 1][par]
        # estimate bb_ns1_orig and bb_ns2_orig using the already computed
        # frontiers and the value of ymargin_bb, following the same approach
        # employed in Slitlet2dArc.__init__()
        poly_lower_frontier = np.polynomial.Polynomial(
            outdict['contents'][cslitlet]['frontier']['poly_coef_lower']
        )
        poly_upper_frontier = np.polynomial.Polynomial(
            outdict['contents'][cslitlet]['frontier']['poly_coef_upper']
        )
        ylower = poly_lower_frontier(xdum)
        yupper = poly_upper_frontier(xdum)
        ymargin_bb = master_rectwv.contents[islitlet - 1]['ymargin_bb']
        bb_ns1_orig = int(ylower.min() + 0.5) - ymargin_bb
        if bb_ns1_orig < 1:
            bb_ns1_orig = 1
        bb_ns2_orig = int(yupper.max() + 0.5) + ymargin_bb
        if bb_ns2_orig > EMIR_NAXIS2:
            bb_ns2_orig = EMIR_NAXIS2
        outdict['contents'][cslitlet]['bb_ns1_orig'] = bb_ns1_orig
        outdict['contents'][cslitlet]['bb_ns2_orig'] = bb_ns2_orig

    # additional parameters (see Slitlet2dArc.__init__)
    for islitlet in list_valid_islitlets:
        cslitlet = 'slitlet' + str(islitlet).zfill(2)
        # define expected frontier ordinates at x0_reference for the rectified
        # image imposing the vertical length of the slitlet to be constant
        # and equal to EMIR_NPIXPERSLIT_RECTIFIED
        outdict['contents'][cslitlet]['y0_frontier_lower_expected'] = \
            expected_y0_lower_frontier(islitlet)
        outdict['contents'][cslitlet]['y0_frontier_upper_expected'] = \
            expected_y0_upper_frontier(islitlet)
        # compute linear transformation to place the rectified slitlet at
        # the center of the current slitlet bounding box
        tmpdict = outdict['contents'][cslitlet]
        xdum1 = tmpdict['y0_frontier_lower']
        ydum1 = tmpdict['y0_frontier_lower_expected']
        xdum2 = tmpdict['y0_frontier_upper']
        ydum2 = tmpdict['y0_frontier_upper_expected']
        corr_yrect_b = (ydum2 - ydum1) / (xdum2 - xdum1)
        corr_yrect_a = ydum1 - corr_yrect_b * xdum1
        # compute expected location of rectified boundaries
        y0_reference_lower_expected = \
            corr_yrect_a + corr_yrect_b * tmpdict['y0_reference_lower']
        y0_reference_middle_expected = \
            corr_yrect_a + corr_yrect_b * tmpdict['y0_reference_middle']
        y0_reference_upper_expected = \
            corr_yrect_a + corr_yrect_b * tmpdict['y0_reference_upper']
        # shift transformation to center the rectified slitlet within the
        # slitlet bounding box
        ydummid = (ydum1 + ydum2) / 2
        ioffset = int(
            ydummid - (tmpdict['bb_ns1_orig'] + tmpdict['bb_ns2_orig']) / 2.0)
        corr_yrect_a -= ioffset
        # minimum and maximum row in the rectified slitlet encompassing
        # EMIR_NPIXPERSLIT_RECTIFIED pixels
        # a) scan number (in pixels, from 1 to NAXIS2)
        xdum1 = corr_yrect_a + \
                corr_yrect_b * tmpdict['y0_frontier_lower']
        xdum2 = corr_yrect_a + \
                corr_yrect_b * tmpdict['y0_frontier_upper']
        # b) row number (starting from zero)
        min_row_rectified = \
            int((round(xdum1 * 10) + 5) / 10) - tmpdict['bb_ns1_orig']
        max_row_rectified = \
            int((round(xdum2 * 10) - 5) / 10) - tmpdict['bb_ns1_orig']
        # save previous results in outdict
        outdict['contents'][cslitlet]['y0_reference_lower_expected'] = \
            y0_reference_lower_expected
        outdict['contents'][cslitlet]['y0_reference_middle_expected'] = \
            y0_reference_middle_expected
        outdict['contents'][cslitlet]['y0_reference_upper_expected'] = \
            y0_reference_upper_expected
        outdict['contents'][cslitlet]['corr_yrect_a'] = corr_yrect_a
        outdict['contents'][cslitlet]['corr_yrect_b'] = corr_yrect_b
        outdict['contents'][cslitlet]['min_row_rectified'] = min_row_rectified
        outdict['contents'][cslitlet]['max_row_rectified'] = max_row_rectified

    # ---

    # Create object of type RectWaveCoeff with coefficients for
    # rectification and wavelength calibration
    rectwv_coeff = RectWaveCoeff(instrument='EMIR')
    rectwv_coeff.quality_control = numina.types.qc.QC.GOOD
    rectwv_coeff.tags['grism'] = grism_name
    rectwv_coeff.tags['filter'] = filter_name
    rectwv_coeff.meta_info['origin']['bound_param'] = \
        master_rectwv.meta_info['origin']['bound_param']
    rectwv_coeff.meta_info['origin']['master_rectwv'] = \
        'uuid' + master_rectwv.uuid
    rectwv_coeff.meta_info['dtu_configuration'] = outdict['dtu_configuration']
    rectwv_coeff.total_slitlets = EMIR_NBARS
    for i in range(EMIR_NBARS):
        islitlet = i + 1
        dumdict = {'islitlet': islitlet}
        cslitlet = 'slitlet' + str(islitlet).zfill(2)
        if cslitlet in outdict['contents']:
            dumdict.update(outdict['contents'][cslitlet])
        else:
            dumdict.update({
                'csu_bar_left': csu_conf.csu_bar_left(islitlet),
                'csu_bar_right': csu_conf.csu_bar_right(islitlet),
                'csu_bar_slit_center': csu_conf.csu_bar_slit_center(islitlet),
                'csu_bar_slit_width': csu_conf.csu_bar_slit_width(islitlet),
                'x0_reference': float(EMIR_NAXIS1) / 2.0 + 0.5,
                'y0_frontier_lower_expected':
                    expected_y0_lower_frontier(islitlet),
                'y0_frontier_upper_expected':
                    expected_y0_upper_frontier(islitlet)
            })
            rectwv_coeff.missing_slitlets.append(islitlet)
        rectwv_coeff.contents.append(dumdict)
    # debugging __getstate__ and __setstate__
    # rectwv_coeff.writeto(args.out_rect_wpoly.name)
    # print('>>> Saving file ' + args.out_rect_wpoly.name)
    # check_setstate_getstate(rectwv_coeff, args.out_rect_wpoly.name)
    logger.info('Generating RectWaveCoeff object with uuid=' +
                rectwv_coeff.uuid)

    return rectwv_coeff


def main(args=None):
    # parse command-line options
    parser = argparse.ArgumentParser(
        description='description: evaluate rectification and wavelength '
                    'calibration polynomials for the CSU configuration of a '
                    'particular image'
    )

    # required arguments
    parser.add_argument("fitsfile",
                        help="Input FITS file",
                        type=argparse.FileType('rb'))
    parser.add_argument("--rect_wpoly_MOSlibrary", required=True,
                        help="Input JSON file with library of rectification "
                             "and wavelength calibration coefficients",
                        type=argparse.FileType('rt'))
    parser.add_argument("--out_json", required=True,
                        help="Output JSON file with calibration computed for "
                             "the input FITS file",
                        type=lambda x: arg_file_is_new(parser, x, mode='wt'))
    # optional arguments
    parser.add_argument("--global_integer_offset_x_pix",
                        help="Global integer offset in the X direction "
                             "(default=0)",
                        default=0, type=int)
    parser.add_argument("--global_integer_offset_y_pix",
                        help="Global integer offset in the Y direction "
                             "(default=0)",
                        default=0, type=int)
    parser.add_argument("--ignore_dtu_configuration",
                        help="Ignore DTU configurations differences between "
                             "model and input image",
                        action="store_true")
    parser.add_argument("--debugplot",
                        help="Integer indicating plotting & debugging options"
                             " (default=0)",
                        default=0, type=int,
                        choices=DEBUGPLOT_CODES)
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")
    args = parser.parse_args(args)

    if args.echo:
        print('\033[1m\033[31m% ' + ' '.join(sys.argv) + '\033[0m\n')

    # ---

    # generate HDUList object
    hdulist = fits.open(args.fitsfile)

    # generate MasterRectWave object
    master_rectwv = MasterRectWave._datatype_load(
        args.rect_wpoly_MOSlibrary.name)

    # compute rectification and wavelength calibration coefficients
    rectwv_coeff = rectwv_coeff_from_mos_library(
        hdulist,
        master_rectwv,
        ignore_dtu_configuration=args.ignore_dtu_configuration,
        debugplot=args.debugplot
    )

    # set global offsets
    rectwv_coeff.global_integer_offset_x_pix = \
        args.global_integer_offset_x_pix
    rectwv_coeff.global_integer_offset_y_pix = \
        args.global_integer_offset_y_pix

    # save RectWaveCoeff object into JSON file
    rectwv_coeff.writeto(args.out_json.name)
    print('>>> Saving file ' + args.out_json.name)
    # debugging __getstate__ and __setstate__
    # check_setstate_getstate(rectwv_coeff, args.out_json.name)


if __name__ == "__main__":
    main()
